/*
 *  $Id: isl_oid.h 280 2003-12-26 16:32:31Z hvr $
 *  
 *  Copyright (C) 2003 Herbert Valerio Riedel <hvr@gnu.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#if !defined(_ISL_OID_H)
#define _ISL_OID_H

/* 
 * MIB related constant and structure definitions for communicating
 * with the device firmware
 */

struct obj_ssid {
	u8 length;
	char octets[33];
} __attribute__((packed));

struct obj_key {
	u8 type; /* dot11_priv_t */
	u8 length;
	char key[32];
} __attribute__((packed));

struct obj_mlme {
	u8 address[6];
	u16 id;
	u16 state;
	u16 code;
} __attribute__((packed));

struct obj_mlmeex {
	u8 address[6];
	u16 id;
	u16 state;
	u16 code;
	u16 size;
	u8 data[0];
} __attribute__((packed));

struct obj_buffer {
	u32 size;
	u32 addr;		/* 32bit bus address */
} __attribute__((packed));

struct obj_bss {
	u8 address[6];
	int : 16; /* padding */

	char state;
	char reserved;
	short age;

	char quality;
	char rssi;

	struct obj_ssid ssid;
	short channel;
	char beacon_period;
	char dtim_period;
	short capinfo;
	short rates;
	short basic_rates;
	int : 16; /* padding */
} __attribute__((packed));

struct obj_bsslist {
	u32 nr;
	struct obj_bss bsslist[0];
} __attribute__((packed));

struct obj_frequencies {
	u16 nr;
	u16 mhz[0];
} __attribute__((packed));

/* 
 * in case everything's ok, the inlined function below will be
 * optimized away by the compiler...
 */
static inline void 
__bug_on_wrong_struct_sizes (void)
{
	BUG_ON(sizeof(struct obj_ssid) != 34);
	BUG_ON(sizeof(struct obj_key) != 34);
	BUG_ON(sizeof(struct obj_mlme) != 12);
	BUG_ON(sizeof(struct obj_mlmeex) != 14);
	BUG_ON(sizeof(struct obj_buffer) != 8);
	BUG_ON(sizeof(struct obj_bss) != 60);
	BUG_ON(sizeof(struct obj_bsslist) != 4);
	BUG_ON(sizeof(struct obj_frequencies) != 2);
}

enum oid_t {
	GEN_OID_MACADDRESS		= 0x00000000,
	GEN_OID_LINKSTATE               = 0x00000001,
	GEN_OID_WATCHDOG                = 0x00000002,
	GEN_OID_MIBOP                   = 0x00000003,
	GEN_OID_OPTIONS                 = 0x00000004,
	GEN_OID_LEDCONFIG               = 0x00000005,

	/* 802.11 */
	DOT11_OID_BSSTYPE               = 0x10000000,
	DOT11_OID_BSSID                 = 0x10000001,
	DOT11_OID_SSID                  = 0x10000002,
	DOT11_OID_STATE                 = 0x10000003,
	DOT11_OID_AID                   = 0x10000004,
	DOT11_OID_COUNTRYSTRING         = 0x10000005,
	DOT11_OID_SSIDOVERRIDE          = 0x10000006,

	DOT11_OID_MEDIUMLIMIT           = 0x11000000,
	DOT11_OID_BEACONPERIOD          = 0x11000001,
	DOT11_OID_DTIMPERIOD            = 0x11000002,
	DOT11_OID_ATIMWINDOW            = 0x11000003,
	DOT11_OID_LISTENINTERVAL        = 0x11000004,
	DOT11_OID_CFPPERIOD             = 0x11000005,
	DOT11_OID_CFPDURATION           = 0x11000006,

	DOT11_OID_AUTHENABLE            = 0x12000000,
	DOT11_OID_PRIVACYINVOKED        = 0x12000001,
	DOT11_OID_EXUNENCRYPTED         = 0x12000002,
	DOT11_OID_DEFKEYID              = 0x12000003,
	DOT11_OID_DEFKEY1               = 0x12000004,
	DOT11_OID_DEFKEY2               = 0x12000005,
	DOT11_OID_DEFKEY3               = 0x12000006,
	DOT11_OID_DEFKEY4               = 0x12000007,
	DOT11_OID_STAKEY                = 0x12000008,
	DOT11_OID_REKEYTHRESHOLD        = 0x12000009,
	DOT11_OID_STASC                 = 0x1200000a,

	DOT11_OID_PRIVTXREJECTED        = 0x1a000000,
	DOT11_OID_PRIVRXPLAIN           = 0x1a000001,
	DOT11_OID_PRIVRXFAILED          = 0x1a000002,
	DOT11_OID_PRIVRXNOKEY           = 0x1a000003,

	DOT11_OID_RTSTHRESH             = 0x13000000,
	DOT11_OID_FRAGTHRESH            = 0x13000001,
	DOT11_OID_SHORTRETRIES          = 0x13000002,
	DOT11_OID_LONGRETRIES           = 0x13000003,
	DOT11_OID_MAXTXLIFETIME         = 0x13000004,
	DOT11_OID_MAXRXLIFETIME         = 0x13000005,
	DOT11_OID_AUTHRESPTIMEOUT       = 0x13000006,
	DOT11_OID_ASSOCRESPTIMEOUT      = 0x13000007,

	DOT11_OID_ALOFT_TABLE           = 0x1d000000,
	DOT11_OID_ALOFT_CTRL_TABLE      = 0x1d000001,
	DOT11_OID_ALOFT_RETREAT         = 0x1d000002,
	DOT11_OID_ALOFT_PROGRESS        = 0x1d000003,
	DOT11_OID_ALOFT_FIXEDRATE       = 0x1d000004,
	DOT11_OID_ALOFT_RSSIGRAPH       = 0x1d000005,
	DOT11_OID_ALOFT_CONFIG          = 0x1d000006,

	DOT11_OID_VDCF0                 = 0x1b000000,
	DOT11_OID_VDCF1                 = 0x1b000001,
	DOT11_OID_VDCF2                 = 0x1b000002,
	DOT11_OID_VDCF3                 = 0x1b000003,
	DOT11_OID_VDCF4                 = 0x1b000004,
	DOT11_OID_VDCF5                 = 0x1b000005,
	DOT11_OID_VDCF6                 = 0x1b000006,
	DOT11_OID_VDCF7                 = 0x1b000007,
	DOT11_OID_MAXFRAMEBURST         = 0x1b000008,

	DOT11_OID_PSM                   = 0x14000000,
	DOT11_OID_CAMTIMEOUT            = 0x14000001,
	DOT11_OID_RECEIVEDTIMS          = 0x14000002,
	DOT11_OID_ROAMPREFERENCE        = 0x14000003,

	DOT11_OID_BRIDGELOCAL           = 0x15000000,
	DOT11_OID_CLIENTS               = 0x15000001,
	DOT11_OID_CLIENTSASSOCIATED     = 0x15000002,
	DOT11_OID_CLIENT1               = 0x15000003,
	/* ... */
	DOT11_OID_CLIENT2007            = 0x150007D9,

	DOT11_OID_CLIENTFIND            = 0x150007DB,
	DOT11_OID_WDSLINKADD            = 0x150007DC,
	DOT11_OID_WDSLINKREMOVE         = 0x150007DD,
	DOT11_OID_EAPAUTHSTA		= 0x150007DE,
	DOT11_OID_EAPUNAUTHSTA		= 0x150007DF,
	DOT11_OID_DOT1XENABLE           = 0x150007E0,
	DOT11_OID_MICFAILURE            = 0x150007E1,
	DOT11_OID_REKEYINDICATE         = 0x150007E2,

	DOT11_OID_MPDUTXSUCCESSFUL      = 0x16000000,
	DOT11_OID_MPDUTXONERETRY        = 0x16000001,
	DOT11_OID_MPDUTXMULTIPLERETRIES = 0x16000002,
	DOT11_OID_MPDUTXFAILED          = 0x16000003,
	DOT11_OID_MPDURXSUCCESSFUL      = 0x16000004,
	DOT11_OID_MPDURXDUPS            = 0x16000005,
	DOT11_OID_RTSSUCCESSFUL         = 0x16000006,
	DOT11_OID_RTSFAILED             = 0x16000007,
	DOT11_OID_ACKFAILED             = 0x16000008,
	DOT11_OID_FRAMERECEIVES         = 0x16000009,
	DOT11_OID_FRAMEERRORS           = 0x1600000a,
	DOT11_OID_FRAMEABORTS           = 0x1600000b,
	DOT11_OID_FRAMEABORTSPHY        = 0x1600000c,

	DOT11_OID_SLOTTIME              = 0x17000000,
	DOT11_OID_CWMIN                 = 0x17000001,
	DOT11_OID_CWMAX                 = 0x17000002,
	DOT11_OID_ACKWINDOW             = 0x17000003,
	DOT11_OID_ANTENNARX             = 0x17000004,
	DOT11_OID_ANTENNATX             = 0x17000005,
	DOT11_OID_ANTENNADIVERSITY      = 0x17000006,
	DOT11_OID_CHANNEL               = 0x17000007,
	DOT11_OID_EDTHRESHOLD           = 0x17000008,
	DOT11_OID_PREAMBLESETTINGS      = 0x17000009,
	DOT11_OID_RATES                 = 0x1700000a,
	DOT11_OID_CCAMODESUPPORTED      = 0x1700000b,
	DOT11_OID_CCAMODE               = 0x1700000c,
	DOT11_OID_RSSIVECTOR            = 0x1700000d,
	DOT11_OID_OUTPUTPOWERTABLE      = 0x1700000e,
	DOT11_OID_OUTPUTPOWER           = 0x1700000f,
	DOT11_OID_SUPPORTEDRATES        = 0x17000010,
	DOT11_OID_FREQUENCY             = 0x17000011,
	DOT11_OID_SUPPORTEDFREQUENCIES  = 0x17000012,
	DOT11_OID_NOISEFLOOR            = 0x17000013,
	DOT11_OID_FREQUENCYACTIVITY     = 0x17000014,
	DOT11_OID_IQCALIBRATIONTABLE	= 0x17000015,
	DOT11_OID_NONERPPROTECTION      = 0x17000016,
	DOT11_OID_SLOTSETTINGS          = 0x17000017,
	DOT11_OID_NONERPTIMEOUT         = 0x17000018,
	DOT11_OID_PROFILES		= 0x17000019,
	DOT11_OID_EXTENDEDRATES         = 0x17000020,

	DOT11_OID_DEAUTHENTICATE        = 0x18000000,
	DOT11_OID_AUTHENTICATE          = 0x18000001,
	DOT11_OID_DISASSOCIATE          = 0x18000002,
	DOT11_OID_ASSOCIATE             = 0x18000003,
	DOT11_OID_SCAN                  = 0x18000004,
	DOT11_OID_BEACON                = 0x18000005,
	DOT11_OID_PROBE                 = 0x18000006,
	DOT11_OID_DEAUTHENTICATEEX      = 0x18000007,
	DOT11_OID_AUTHENTICATEEX        = 0x18000008,
	DOT11_OID_DISASSOCIATEEX        = 0x18000009,
	DOT11_OID_ASSOCIATEEX           = 0x1800000a,
	DOT11_OID_REASSOCIATE           = 0x1800000b,
	DOT11_OID_REASSOCIATEEX         = 0x1800000c,

	DOT11_OID_NONERPSTATUS          = 0x1e000000,

	DOT11_OID_STATIMEOUT            = 0x19000000,
	DOT11_OID_MLMEAUTOLEVEL         = 0x19000001,
	DOT11_OID_BSSTIMEOUT            = 0x19000002,
	DOT11_OID_ATTACHMENT            = 0x19000003,
	DOT11_OID_PSMBUFFER             = 0x19000004,

	DOT11_OID_BSSS                  = 0x1c000000,
	DOT11_OID_BSS1                  = 0x1c000001,
	/* .. */
	DOT11_OID_BSS64                 = 0x1c000040,
	DOT11_OID_BSSFIND               = 0x1c000042,
	DOT11_OID_BSSLIST               = 0x1c000043
};

enum dot11_state_t {
	DOT11_STATE_NONE                = 0,
	DOT11_STATE_AUTHING             = 1,
	DOT11_STATE_AUTH                = 2,
	DOT11_STATE_ASSOCING            = 3,
	
	DOT11_STATE_ASSOC               = 5,
	DOT11_STATE_IBSS                = 6,
	DOT11_STATE_WDS			= 7
};

enum dot11_bsstype_t {
	DOT11_BSSTYPE_NONE              = 0,
	DOT11_BSSTYPE_INFRA             = 1,
	DOT11_BSSTYPE_IBSS              = 2,
	DOT11_BSSTYPE_ANY               = 3
};

enum dot11_auth_t {
	DOT11_AUTH_NONE                 = 0,
	DOT11_AUTH_OS                   = 1,
	DOT11_AUTH_SK                   = 2,
	DOT11_AUTH_BOTH                 = 3
};

enum dot11_mlme_t {
	DOT11_MLME_AUTO                 = 0,
	DOT11_MLME_INTERMEDIATE         = 1,
	DOT11_MLME_EXTENDED             = 2
};


enum dot11_priv_t {
	DOT11_PRIV_WEP                  = 0,
	DOT11_PRIV_TKIP			= 1
};

#endif /* !defined(_ISL_OID_H) */
/* EOF */
