/*  $Header: /var/lib/cvs/prism54-ng/ksrc/islpci_dev.h,v 1.39 2004/01/13 21:24:05 jmaurer Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc. 
 *  Copyright (C) 2003 Herbert Valerio Riedel <hvr@gnu.org>
 *  Copyright (C) 2003 Luis R. Rodriguez <mcgrof@ruslug.rutgers.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _ISLPCI_DEV_H
#define _ISLPCI_DEV_H

#include <linux/version.h>
#include <linux/netdevice.h>
#include <linux/wireless.h>
#include <linux/list.h>

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,41)
# include <linux/workqueue.h>
#else
# include <linux/tqueue.h>
# define work_struct tq_struct
# define INIT_WORK INIT_TQUEUE
# define schedule_work schedule_task
#endif

#include "isl_38xx.h"
#include "isl_oid.h"

#ifdef CONFIG_PRISM54_WDS
# include "isl_wds.h"
#endif

typedef struct {
	struct list_head _list;
	u32 host_address;	/* mem address for the host */
	u32 dev_address;	/* mem address for the device */
	u16 size;		/* size of memory block or queue */
	u16 fragments;		/* nr of fragments in data part */
} islpci_qentry_t;

typedef struct {
	struct list_head list;
	int size;		/* size of queue */
	spinlock_t _slock;	/* mutex */
} islpci_queue_t;

/* MIB mirror backup */
struct islpci_mib {
	struct obj_ssid ssid;
	struct obj_key defkey[4];
	char bssid[6];
	char macaddress[6];
	u32 defkeyid;
	u32 config;
	u32 mode;
	u32 iw_mode;
	u32 frequency;
	u32 channel;
	u32 mlmeautolevel;
	u32 authenable;
	u32 privacyinvoked;
	u32 bsstype;
	u32 exunencrypted;
	u32 conformance;
	u32 dot1xenable;
	u32 beaconperiod;
	u32 edthreshold;
	u32 fixedrate;
	u32 rtsthresh;
	u32 fragthresh;
	u32 shortretries;
	u32 longretries;
	u32 maxtxlifetime;
	u32 fixedtxpower;

	struct semaphore sem;
};

/* some states might not be superflous and may be removed when
   design is finalized (hvr) */
typedef enum {
	PRV_STATE_OFF = 0,	/* this means hw_unavailable is != 0 */
	PRV_STATE_PREBOOT,	/* we are in a pre-boot state (empty RAM) */
	PRV_STATE_BOOT,		/* boot state (fw upload, run fw) */
	PRV_STATE_POSTBOOT,	/* after boot state, need reset now */
	PRV_STATE_PREINIT,	/* pre-init state */
	PRV_STATE_INIT,		/* init state (restore MIB backup to device) */
	PRV_STATE_READY,	/* driver&device are in operational state */
	PRV_STATE_SLEEP		/* device in sleep mode */
} islpci_state_t;

/* ACL using MAC address */
struct mac_entry {
   struct list_head _list;
   char addr[ETH_ALEN];
};

struct islpci_acl {
   enum { MAC_POLICY_OPEN=0, MAC_POLICY_ACCEPT=1, MAC_POLICY_REJECT=2 } policy;
   struct list_head mac_list;  /* a list of mac_entry */
   int size;   /* size of queue */
   struct semaphore sem;   /* accessed in ioctls and trap_work */
};

#define MAX_BSS_WPA_IE_COUNT 64
#define MAX_WPA_IE_LEN 64
struct islpci_bss_wpa_ie {
	struct list_head list;
	unsigned long last_update;
	u8 bssid[ETH_ALEN];
	u8 wpa_ie[MAX_WPA_IE_LEN];
	size_t wpa_ie_len;
};

typedef struct {
	spinlock_t slock;	/* generic spinlock; */

	/* prism54_mib_init */
	struct islpci_mib mib;

	/* this job will take care of our wireless stats */
	struct work_struct stats_work;
	struct semaphore stats_sem;

	/* remember when we last updated the stats */
	unsigned long stats_timestamp;

	/* The first is accessed under semaphore locking.
	 * The second is the clean one we return to iwconfig.
	 */
	struct iw_statistics local_iwstatistics;
	struct iw_statistics iwstatistics;

	/* this job will process the traps the card sends us */
	struct work_struct trap_work;

	struct islpci_acl acl;

	/* PCI bus allocation & configuration members */
	struct pci_dev *pdev;	/* PCI structure information */
	u32 pci_state[16];	/* used for suspend/resume */
	char firmware[33];

	void *device_base;	/* ioremapped device base address */

	/* consistent DMA region */
	void *driver_mem_address;	/* base DMA address */
	dma_addr_t device_host_address;	/* base DMA address (bus address) */
	dma_addr_t device_psm_buffer;	/* host memory for PSM buffering (bus address) */

	/* our network_device structure  */
	struct net_device *ndev;

	/* device queue interface members */
	struct isl38xx_cb *control_block;	/* device control block 
							   (== driver_mem_address!) */

	u32 index_mgmt_rx;	/* real index mgmt rx queue */
	u32 index_mgmt_tx;	/* read index mgmt tx queue */
	u32 free_data_rx;	/* free pointer data rx queue */
	u32 free_data_tx;	/* free pointer data tx queue */
	u32 data_low_tx_full;	/* full detected flag */

	/* root free queue definitions */
	islpci_qentry_t _queue_slots[MGMT_FRAME_COUNT];	/* all available queue slots */

	islpci_queue_t mgmt_tx_freeq;	/* queue entry container mgmt tx */
	islpci_queue_t mgmt_rx_freeq;	/* queue entry container mgmt rx */

	/* driver queue definitions */
	islpci_queue_t mgmt_tx_shadowq;	/* shadow queue mgmt tx */
	islpci_queue_t mgmt_rx_shadowq;	/* shadow queue mgmt rx */
	islpci_queue_t ioctl_rx_queue;	/* ioctl rx queue */
	islpci_queue_t trap_rx_queue;	/* TRAP pimfor rx queue */
	islpci_queue_t pimfor_rx_queue;	/* PIMFOR generated rx queue */

	struct sk_buff *data_low_tx[ISL38XX_CB_TX_QSIZE];
	struct sk_buff *data_low_rx[ISL38XX_CB_RX_QSIZE];
	dma_addr_t pci_map_tx_address[ISL38XX_CB_TX_QSIZE];
	dma_addr_t pci_map_rx_address[ISL38XX_CB_RX_QSIZE];

	/* driver network interface members */
	struct net_device_stats statistics;

	/* wait for a reset interrupt */
	wait_queue_head_t reset_done;

	/* used by islpci_rec islpci_mgt_receive/pci_mgt_response */
	wait_queue_head_t ioctl_reader_wqueue;
	struct semaphore ioctl_reader_sem;
	atomic_t ioctl_reader_count;

	/* state machine */
	islpci_state_t state;
	int state_off;		/* enumeration of off-state, if 0 then
				 * we're not in any off-state */

	int wpa; /* WPA mode enabled */
	struct list_head bss_wpa_list;
	int num_bss_wpa;

#ifdef CONFIG_PRISM54_WDS
	struct wds_priv wdsp;
#endif
} islpci_private;

static inline islpci_state_t
islpci_get_state(islpci_private *priv)
{
	/* lock */
	return priv->state;
	/* unlock */
}

static inline islpci_state_t
islpci_set_state(islpci_private *priv, islpci_state_t new_state)
{
	islpci_state_t old_state;

	/* lock */
	old_state = priv->state;

	/* this means either a race condition or some serious error in
	 * the driver code */
	switch (new_state) {
	case PRV_STATE_OFF:
		priv->state_off++;
	default:
		priv->state = new_state;
		break;

	case PRV_STATE_PREBOOT:
		/* there are actually many off-states, enumerated by
		 * state_off */
		if (old_state == PRV_STATE_OFF)
			priv->state_off--;

		/* only if hw_unavailable is zero now it means we either
		 * were in off#1 state, or came here from
		 * somewhere else */
		if (!priv->state_off)
			priv->state = new_state;
		break;
	};
#if 0
	printk(KERN_DEBUG "%s: state transition %d -> %d (off#%d)\n",
	       priv->ndev->name, old_state, new_state, priv->state_off);
#endif

	/* invariants */
	BUG_ON(priv->state_off < 0);
	BUG_ON(priv->state_off && (priv->state != PRV_STATE_OFF));
	BUG_ON(!priv->state_off && (priv->state == PRV_STATE_OFF));

	/* unlock */
	return old_state;
}

#define ISLPCI_TX_TIMEOUT               (2*HZ)

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,75))
# define irqreturn_t void
# define IRQ_HANDLED
# define IRQ_NONE
#endif

irqreturn_t islpci_interrupt(int, void *, struct pt_regs *);

int prism54_post_setup(islpci_private *, int);
int islpci_reset(islpci_private *, int);

static inline void
islpci_trigger(islpci_private *priv)
{
	isl38xx_trigger_device(islpci_get_state(priv) == PRV_STATE_SLEEP,
			       priv->device_base);
}

struct net_device_stats *islpci_statistics(struct net_device *);

int islpci_alloc_memory(islpci_private *);
int islpci_free_memory(islpci_private *);
struct net_device *islpci_setup(struct pci_dev *);

#endif				/* _ISLPCI_DEV_H */
