/*  $Header: /var/lib/cvs/prism54-ng/ksrc/islpci_mgt.h,v 1.19 2004/01/09 15:35:27 mcgrof Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *  Copyright (C) 2003 Luis R. Rodriguez <mcgrof@ruslug.rutgers.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _ISLPCI_MGT_H
#define _ISLPCI_MGT_H

#include <linux/wireless.h>
#include <linux/skbuff.h>

#include "islpci_dev.h"

/*
 *  Function definitions
 */

#define K_DEBUG(f, m, args...) do { if(f & m) printk(KERN_DEBUG args); } while(0)
#define DEBUG(f, args...) K_DEBUG(f, pc_debug, args)

#define TRACE(devname)   K_DEBUG(SHOW_TRACING, VERBOSE, "%s:  -> " __FUNCTION__ "()\n", devname)

extern int pc_debug;
#ifdef CONFIG_PRISM54_WDS
extern int init_wds;
#else
static const int init_wds = 0;	/* help compiler optimize away dead code */
#endif


/* General driver definitions */
#define PCIVENDOR_INTERSIL                      0x1260UL
#define PCIVENDOR_3COM				0x10b7UL
#define PCIVENDOR_DLINK				0x1186UL
#define PCIVENDOR_I4				0x17cfUL
#define PCIVENDOR_IODATA			0x10fcUL
#define PCIVENDOR_NETGEAR			0x1385UL
#define PCIVENDOR_SMC				0x10b8UL

#define PCIDEVICE_ISL3877                       0x3877UL
#define PCIDEVICE_ISL3890                       0x3890UL
#define	PCIDEVICE_3COM6001			0x6001UL
#define PCIDEVICE_LATENCY_TIMER_MIN 		0x40
#define PCIDEVICE_LATENCY_TIMER_VAL 		0x50

/* Debugging verbose definitions */
#define SHOW_NOTHING                            0x00	/* overrules everything */
#define SHOW_ANYTHING                           0xFF
#define SHOW_ERROR_MESSAGES                     0x01
#define SHOW_TRAPS                              0x02
#define SHOW_FUNCTION_CALLS                     0x04
#define SHOW_TRACING                            0x08
#define SHOW_QUEUE_INDEXES                      0x10
#define SHOW_PIMFOR_FRAMES                      0x20
#define SHOW_BUFFER_CONTENTS                    0x40
#define VERBOSE                                 0x01

/* Default card definitions */
#define CARD_DEFAULT_CHANNEL                    6
#define CARD_DEFAULT_MODE                       INL_MODE_CLIENT
#define CARD_DEFAULT_IW_MODE			IW_MODE_INFRA
#define CARD_DEFAULT_BSSTYPE                    DOT11_BSSTYPE_INFRA
#define CARD_DEFAULT_CLIENT_SSID		""
#define CARD_DEFAULT_AP_SSID			"default"
#define CARD_DEFAULT_KEY1                       "default_key_1"
#define CARD_DEFAULT_KEY2                       "default_key_2"
#define CARD_DEFAULT_KEY3                       "default_key_3"
#define CARD_DEFAULT_KEY4                       "default_key_4"
#define CARD_DEFAULT_WEP                        0
#define CARD_DEFAULT_FILTER                     0
#ifdef CONFIG_PRISM54_WDS
# define CARD_DEFAULT_WDS                        1
#else
# define CARD_DEFAULT_WDS                        0
#endif
#define	CARD_DEFAULT_AUTHEN                     DOT11_AUTH_OS
#define	CARD_DEFAULT_DOT1X			0
#define CARD_DEFAULT_MLME_MODE			DOT11_MLME_AUTO
#define CARD_DEFAULT_CONFORMANCE                OID_INL_CONFORMANCE_NONE

/* PIMFOR package definitions */
#define PIMFOR_ETHERTYPE                        0x8828
#define PIMFOR_HEADER_SIZE                      12
#define PIMFOR_VERSION                          1
#define PIMFOR_OP_GET                           0
#define PIMFOR_OP_SET                           1
#define PIMFOR_OP_RESPONSE                      2
#define PIMFOR_OP_ERROR                         3
#define PIMFOR_OP_TRAP                          4
#define PIMFOR_OP_RESERVED                      5	/* till 255 */
#define PIMFOR_DEV_ID_MHLI_MIB                  0
#define PIMFOR_FLAG_APPLIC_ORIGIN               0x01
#define PIMFOR_FLAG_LITTLE_ENDIAN               0x02

/* Driver specific oid definitions (NDIS driver) */
#define OID_INL_TUNNEL                          0xFF020000
#define OID_INL_MEMADDR                         0xFF020001
#define OID_INL_MEMORY                          0xFF020002
#define OID_INL_MODE                            0xFF020003
#define OID_INL_COMPONENT_NR                    0xFF020004
#define OID_INL_VERSION                         0xFF020005
#define OID_INL_INTERFACE_ID                    0xFF020006
#define OID_INL_COMPONENT_ID                    0xFF020007
#define OID_INL_CONFIG                          0xFF020008
#define OID_INL_DOT11D_CONFORMANCE              0xFF02000C
#define OID_INL_PHYCAPABILITIES					0xFF02000D
#define	OID_INL_OUTPUTPOWER 					0xFF02000F


/* The dot11d conformance level configures the 802.11d conformance levels.
 * The following conformance levels exist:*/
#define OID_INL_CONFORMANCE_NONE        0	/* Perform active scanning */
#define OID_INL_CONFORMANCE_STRICT      1	/* Strictly adhere to 802.11d */
#define OID_INL_CONFORMANCE_FLEXIBLE    2	/* Use passed 802.11d info to
						 * determine channel AND/OR just make 
						 * assumption that active 
						 * channels are valid  channels */

/* Mode object definitions (OID_INL_MODE) */
#define INL_MODE_NONE                           -1
#define INL_MODE_PROMISCUOUS                    0
#define INL_MODE_CLIENT                         1
#define INL_MODE_AP                             2
#define INL_MODE_SNIFFER                        3

/* Config object definitions (OID_INL_CONFIG) */
#define INL_CONFIG_NOTHING			0x00
#define INL_CONFIG_MANUALRUN                    0x01
#define INL_CONFIG_FRAMETRAP                    0x02
#define INL_CONFIG_RXANNEX                      0x04
#define INL_CONFIG_TXANNEX                      0x08
#define INL_CONFIG_WDS                          0x10

/* Card radio capabilities. Read only bitmask */
#define INL_PHYCAP_2400MHZ			1
#define INL_PHYCAP_5000MHZ			2
#define INL_PHYCAP_FAA				0x80000000 /* Means card supports the FAA switch */

static inline void
add_le32p(u32 * le_number, u32 add)
{
	*le_number = cpu_to_le32(le32_to_cpup(le_number) + add);
}

void display_buffer(char *, int);
void print_frame(struct sk_buff *);
void string_to_macaddress(unsigned char *, unsigned char *);
/*
 *  Type definition section
 *
 *  the structure defines only the header allowing copyless
 *  frame handling
 */
typedef struct {
	u8 version;
	u8 operation;
	u32 oid;
	u8 device_id;
	u8 flags;
	u32 length;
} __attribute__ ((packed))
pimfor_header_t;

void
pimfor_encode_header(int, u32, int, int, u32, pimfor_header_t *);
void
pimfor_decode_header(const pimfor_header_t *, int *, u32 *, int *,
		     int *, u32 *);

void
islpci_queue_init(islpci_queue_t *);
int
islpci_queue_put(islpci_queue_t *, islpci_qentry_t *);
int
islpci_queue_get(islpci_queue_t *, islpci_qentry_t **);
int
islpci_queue_size(islpci_queue_t *);

void
islpci_mgt_initialize(islpci_private *);
int
islpci_mgt_transmit(islpci_private *);
int
islpci_mgt_queue(islpci_private *, int, unsigned long, int, void *, int, int);
int
islpci_mgt_receive(islpci_private *);

int
islpci_mgt_response(islpci_private *, long, int *, void **, long *,
		    islpci_qentry_t **);
void
islpci_mgt_release(islpci_private *, islpci_qentry_t *);

int
islpci_mgt_mlme(unsigned long, char *);

#ifdef CONFIG_PRISM54_WDS
int
islpci_wdslink_add_hndl(unsigned int, unsigned int, unsigned int, unsigned,
			unsigned int, unsigned int, void *, unsigned long);
int
islpci_wdslink_del_hndl(unsigned int, unsigned int, unsigned int, unsigned,
			unsigned int, unsigned int, void *, unsigned long);
#endif

#endif				/* _ISLPCI_MGT_H */

void
islpci_parse_trap(islpci_qentry_t *, u32 *, char **);
