/*  $Header: /var/lib/cvs/prism54-ng/ksrc/isl_ioctl.c,v 1.48 2003/11/13 17:37:59 ajfa Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *            (C) 2003 Aurelien Alleaume <slts@free.fr>
 *            (C) 2003 Herbert Valerio Riedel <hvr@gnu.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/if_arp.h>
#include <linux/pci.h>

#include <asm/uaccess.h>

#include "isl_ioctl.h"
#include "isl_gen.h"
#include "islpci_mgt.h"
#include "bloboid.h"		/* additional types and defs for isl38xx fw */

#undef IW_MAX_BITRATES
#define	IW_MAX_BITRATES	16

#include <linux/wireless.h>

#if WIRELESS_EXT > 12
#include <net/iw_handler.h>	// New driver API
#endif				/* WIRELESS_EXT > 12 */

/****************************************************************************
 * common patterns 
 */

/* set patterns */

static inline int
_mgt_set_request(islpci_private *priv, u32 oid, void *data, unsigned dlen)
{
	int _rvalue, _op = PIMFOR_OP_ERROR;
	unsigned long _dlen;
	void *_data;
	islpci_qentry_t *_entry = 0;

	islpci_mgt_queue(priv, PIMFOR_OP_SET, oid, 0, data, dlen, 0);
	_rvalue = islpci_mgt_response(priv, oid, &_op, &_data, &_dlen, &_entry);
	if (!_rvalue)
		islpci_mgt_release(priv, _entry);
	else
		return _rvalue;

	if (_op == PIMFOR_OP_ERROR)
		return -EINVAL;

	return 0;
}

static inline int
_mgt_set_request_u32(islpci_private *priv, u32 oid, u32 datum)
{
	u32 _datum = cpu_to_le32(datum);
	return _mgt_set_request(priv, oid, &_datum, sizeof (_datum));
}

#define MGT_SET_REQUEST(priv,oid,data,dlen) do { \
	int _rvalue = _mgt_set_request ((priv), (oid), (data), (dlen)); \
	if (_rvalue) return _rvalue; 	\
	} while (0)

#define MGT_SET_REQUEST_U32(priv,oid,datum) do { \
	int _rvalue = _mgt_set_request_u32 ((priv), (oid), (datum)); \
	if (_rvalue) return _rvalue; 	\
	} while (0)

/* get patterns */

static inline int
_mgt_get_request(islpci_private *priv, u32 oid, void *data, void **res,
		 unsigned dlen, islpci_qentry_t **entry)
{
	int _rvalue, _op = PIMFOR_OP_ERROR;
	unsigned long _dlen;

	*entry = NULL;

	islpci_mgt_queue(priv, PIMFOR_OP_GET, oid, 0, data, dlen, 0);

	_rvalue = islpci_mgt_response(priv, oid, &_op, res, &_dlen, entry);

	if (_rvalue)
		return _rvalue;
	if (_op == PIMFOR_OP_ERROR) {
		islpci_mgt_release(priv, *entry);
		return -EINVAL;
	}

	if (_dlen != dlen)
		printk(KERN_DEBUG
		       "_mgt_get_request(): data length (%ld!=%d) mismatch for oid=0x%x\n",
		       _dlen, dlen, oid);

	return 0;
}

/* a version returning what we want in case of failure */
#define MGT_GET_REQUEST_2(priv,oid,data,res,dlen,entry,ret) do { \
	int _rvalue =			\
		_mgt_get_request ((priv), (oid), (data), (res), (dlen), (entry)); \
	if (_rvalue)			\
		return (ret); 	\
	} while (0)

/* version returning _mgt_get_request's retval */
#define MGT_GET_REQUEST(priv,oid,data,res,dlen,entry) \
	MGT_GET_REQUEST_2((priv),(oid),(data),(res),(dlen),(entry),_rvalue)

#define MGT_FREE_REQUEST(priv,entry) do { \
	if ((entry))			\
		islpci_mgt_release((priv), (entry)); 	\
	} while (0)

static inline int
_mgt_get_request_u32(islpci_private *priv, u32 oid, u32 * datum)
{
	u32 *_result = 0;
	islpci_qentry_t *_entry = 0;

	MGT_GET_REQUEST(priv, oid, 0, (void **) &_result, sizeof (u32),
			&_entry);
	*datum = le32_to_cpu(*_result);
	MGT_FREE_REQUEST(priv, _entry);

	return 0;
}

#define MGT_GET_REQUEST_U32_2(priv,oid,datum,ret) do { \
	int _rvalue = _mgt_get_request_u32 ((priv), (oid), (datum)); \
	if (_rvalue) return (ret); 	\
	} while (0)

#define MGT_GET_REQUEST_U32(priv,oid,datum)	\
	MGT_GET_REQUEST_U32_2((priv),(oid),(datum),_rvalue)
	
/****************************************************************************/

int init_mode = CARD_DEFAULT_MODE;
int init_channel = CARD_DEFAULT_CHANNEL;
int init_bsstype = CARD_DEFAULT_BSSTYPE;
int init_wep = CARD_DEFAULT_WEP;
int init_filter = CARD_DEFAULT_FILTER;
int init_authen = CARD_DEFAULT_AUTHEN;
int init_dot1x = CARD_DEFAULT_DOT1X;
int init_conformance = CARD_DEFAULT_CONFORMANCE;
int init_mlme = CARD_DEFAULT_MLME_MODE;

MODULE_PARM(init_mode, "i");
MODULE_PARM(init_channel, "i");
MODULE_PARM(init_bsstype, "i");
MODULE_PARM(init_wep, "i");
MODULE_PARM(init_filter, "i");
MODULE_PARM(init_authen, "i");
MODULE_PARM(init_dot1x, "i");
MODULE_PARM(init_mlme, "i");

#ifdef WDS_LINKS
int init_wds = CARD_DEFAULT_WDS;
MODULE_PARM(init_wds, "i");
#endif

void
prism54_mib_init_defaults(struct islpci_mib *mib)
{
	int idx;

	memset(mib, 0, sizeof *mib);

	mib->mode = init_mode;
	mib->channel = init_channel;
	mib->bsstype = init_bsstype;
	mib->authenable = init_authen;
	mib->privacyinvoked = init_wep;
	mib->exunencrypted = init_filter;
	mib->dot1xenable = init_dot1x;
	mib->mlmeautolevel = init_mlme;
	mib->conformance = init_conformance;

	for (idx = 0; idx < 4; ++idx) {
		mib->defkey[idx].type = DOT11_PRIV_WEP;
		mib->defkey[idx].length = 0;
	}
}

int
prism54_mib_init(islpci_private *priv)
{
	struct obj_buffer psm_buffer = { 
	  .size = cpu_to_le32(PSM_BUFFER_SIZE),
	  .addr = cpu_to_le32(priv->device_psm_buffer)
	};

	printk(KERN_DEBUG "%s: prism54_mib_init()\n", priv->ndev->name);

	/* update ndev's MAC address from firmware */
	{
		void *mac_addr = 0;
		islpci_qentry_t *entry = 0;

		MGT_GET_REQUEST(priv, GEN_OID_MACADDRESS, NULL, &mac_addr, 6,
				&entry);
		memcpy(priv->ndev->dev_addr, mac_addr, 6);
		MGT_FREE_REQUEST(priv, entry);
	}

	/* set the configuration mode to manual run & wds option */
	if (init_wds)
		MGT_SET_REQUEST_U32(priv, OID_INL_CONFIG,
				    INL_CONFIG_MANUALRUN | INL_CONFIG_WDS);
	else
		MGT_SET_REQUEST_U32(priv, OID_INL_CONFIG, INL_CONFIG_MANUALRUN);

	/* set the mode a 1st time for stopping the device */
	MGT_SET_REQUEST_U32(priv, OID_INL_MODE, priv->mib.mode);

	/* set the bsstype */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_BSSTYPE, priv->mib.bsstype);

	/* set the channel/frequency */
	if (priv->mib.frequency)
		MGT_SET_REQUEST_U32(priv, DOT11_OID_FREQUENCY,
				    priv->mib.frequency);
	else /* frequency == 0 -> use channel value */
		MGT_SET_REQUEST_U32(priv, DOT11_OID_CHANNEL,
				    priv->mib.channel);

	/* set the SSID */
	MGT_SET_REQUEST(priv, DOT11_OID_SSID, (void *) &priv->mib.ssid,
			sizeof (struct obj_ssid));

	/* set the PSM buffer object */
	MGT_SET_REQUEST(priv, DOT11_OID_PSMBUFFER,
			(void *) &psm_buffer, sizeof (struct obj_buffer));

	/* set the authentication enable option */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_AUTHENABLE, priv->mib.authenable);

	/* set the privacy invoked mode */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_PRIVACYINVOKED,
			    priv->mib.privacyinvoked);

	/* set the exunencrypted object */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_EXUNENCRYPTED,
			    priv->mib.exunencrypted);

	/* set default key 1-4 */
	MGT_SET_REQUEST(priv, DOT11_OID_DEFKEY1,
			&priv->mib.defkey[0], sizeof (struct obj_key));

	MGT_SET_REQUEST(priv, DOT11_OID_DEFKEY2,
			&priv->mib.defkey[1], sizeof (struct obj_key));

	MGT_SET_REQUEST(priv, DOT11_OID_DEFKEY3,
			&priv->mib.defkey[2], sizeof (struct obj_key));

	MGT_SET_REQUEST(priv, DOT11_OID_DEFKEY4,
			&priv->mib.defkey[3], sizeof (struct obj_key));

	/* set the dot1x enable object */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_DOT1XENABLE, priv->mib.dot1xenable);

	/* set the 802.11d conformance mode object  */
	MGT_SET_REQUEST_U32(priv, OID_INL_DOT11D_CONFORMANCE,
			    priv->mib.conformance);

	/* set the mlme mode object */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_MLMEAUTOLEVEL,
			    priv->mib.mlmeautolevel);

	/* set the mode again for actually starting the device */
	MGT_SET_REQUEST_U32(priv, OID_INL_MODE, priv->mib.mode);

	printk(KERN_DEBUG "%s: done with prism54_mib_init()!\n",
	       priv->ndev->name);

	return 0;
}

struct iw_statistics *
prism54_get_wireless_stats(struct net_device *ndev)
{
	islpci_private *priv = ndev->priv;
	islpci_qentry_t *entry;
	char *data;
	u32 u;
	int j;
	struct obj_bss bss, *bss2;

/* missing stats are :
 *  iwstatistics.qual.updated
 *  iwstatistics.discard.nwid	    
 *  iwstatistics.discard.fragment	    
 *  iwstatistics.discard.misc
 *  iwstatistics.miss.beacon */

/* The value we will return if we fail.
 * eg : if we are not connected to any AP
 */
	priv->iwstatistics.qual.qual = 0;
	priv->iwstatistics.qual.level = 0;
	priv->iwstatistics.qual.noise = 0;

/* Noise floor.
 * I'm not sure if the unit is dBm.
 * Note : If we are not connected, this value seems to be irrevelant. */

	MGT_GET_REQUEST_U32_2(priv, DOT11_OID_NOISEFLOOR, &u, &priv->iwstatistics);
	priv->iwstatistics.qual.noise = u;
	
/* Get the rssi of the link. To do this we need to retrieve a bss.
 * I'm not sure if the unit is dBm. */

	/* First get the MAC address of the AP we are associated with. */
	MGT_GET_REQUEST_2(priv, DOT11_OID_BSSID, NULL,
			  (void **) &data, 6, &entry, &priv->iwstatistics);

	/* copy this MAC to the bss */
	for (j = 0; j < 6; j++)
		bss.address[j] = data[j];
	MGT_FREE_REQUEST(priv, entry);

	/* now ask for the corresponding bss */
	MGT_GET_REQUEST_2(priv, DOT11_OID_BSSFIND, (void *) &bss,
			  (void **) &bss2, sizeof (struct obj_bss), &entry,
			  &priv->iwstatistics);

	/* report the rssi and use it to calculate
	 *  link quality through a signal-noise
	 *  ratio */
	priv->iwstatistics.qual.level = bss2->rssi;
	priv->iwstatistics.qual.qual =
	    bss2->rssi - priv->iwstatistics.qual.noise;

	MGT_FREE_REQUEST(priv, entry);

/* Rx : unable to decrypt the MPDU */
	MGT_GET_REQUEST_U32_2(priv, DOT11_OID_PRIVRXFAILED,
	          &u, &priv->iwstatistics);
	priv->iwstatistics.discard.code = u;
	
/* Tx : Max MAC retries num reached */
	MGT_GET_REQUEST_U32_2(priv, DOT11_OID_MPDUTXFAILED,
			  &u, &priv->iwstatistics);
	priv->iwstatistics.discard.retries = u;
	
	return &priv->iwstatistics;
}

int
prism54_commit(struct net_device *ndev, struct iw_request_info *info,
	       char *cwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	/* printk(KERN_DEBUG "%s: SIOCSIWCOMMIT called\n", ndev->name); */

	/* simply re-set the last set SSID, this should commit most stuff */
	MGT_SET_REQUEST(priv, DOT11_OID_SSID, &priv->mib.ssid,
			sizeof (struct obj_ssid));

	return 0;
}

int
prism54_get_name(struct net_device *ndev, struct iw_request_info *info,
		 char *cwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	if (priv->pdev->device == PCIDEVICE_ISL3877)
		strncpy(cwrq, "PRISM Indigo", IFNAMSIZ);
	else if (priv->pdev->device == PCIDEVICE_ISL3890)
		strncpy(cwrq, "PRISM Duette", IFNAMSIZ);
	else if (priv->pdev->device == PCIDEVICE_3COM6001)
		strncpy(cwrq, "3COM 3CRWE154G72", IFNAMSIZ);
	else
		strncpy(cwrq, "Unsupported card type", IFNAMSIZ);
	return 0;
}

int
prism54_set_freq(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_freq *fwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	long oid;
	u32 channel = 0;
	int mantisse, exponent;

	/* prepare the structure for the set object */
	if (fwrq->m < 1000) {
		/* structure value contains a channel indication */
		oid = DOT11_OID_CHANNEL;
		channel = fwrq->m;
	} else {
		/* structure contains a frequency indication */
		oid = DOT11_OID_FREQUENCY;

		/* mantisse is larger than 1000, so make it kHz */
		mantisse = fwrq->m / 1000;
		exponent = fwrq->e;

		/* process the exponent */
		while (exponent > 0) {
			mantisse *= 10;
			exponent--;
		}

		BUG_ON(exponent != 0);
		channel = mantisse;
	}

	/* request the device to set the channel */
	MGT_SET_REQUEST_U32(priv, oid, channel);

	/* alright, request suceeded, update our MIB mirror */
	priv->mib.frequency = priv->mib.channel = 0;

	if (oid == DOT11_OID_FREQUENCY)
		priv->mib.frequency = channel;
	else
		priv->mib.channel = channel;

	/* Call commit handler */
	return -EINPROGRESS;
}

int
prism54_get_freq(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_freq *fwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 channel;

	/* request the device for the channel */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_FREQUENCY, &channel);

	/* check whether it concern a channel number or a frequency
	 * this becomes difficult because the frequency is in kHz */
	if (channel < 1000) {
		/* interpret the data as a channel */
		fwrq->m = le32_to_cpu(channel);
		fwrq->e = 0;
	} else {
		/* interpret the data as frequency */
		fwrq->m = le32_to_cpu(channel);
		fwrq->e = 3;
	}

	return 0;
}

/* Maybe we should create 2 helpers enter_monitor_mode and leave_monitor_mode
 * to make prism54_set_mode more readable.
 */

int
prism54_set_mode(struct net_device *ndev, struct iw_request_info *info,
		 __u32 * uwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	u32 cardmode = INL_MODE_NONE;
	u32 bsstype = DOT11_BSSTYPE_NONE;
	u32 cardconfig;
	
	/* To report full 802.11 frames in monitor mode we need to enable
 	 * INL_CONFIG_TXANNEX in OID_INL_CONFIG.
	 */
	MGT_GET_REQUEST_U32(priv, OID_INL_CONFIG, &cardconfig);

	switch (*uwrq) {
	case IW_MODE_ADHOC:
		bsstype = DOT11_BSSTYPE_IBSS;
		cardmode = INL_MODE_CLIENT;
		cardconfig &= ~INL_CONFIG_RXANNEX;
		ndev->type = ARPHRD_ETHER;
		clear_bit(FLAG_PROMISC, &priv->flags);
		break;

	case IW_MODE_INFRA:
		bsstype = DOT11_BSSTYPE_INFRA;
		cardmode = INL_MODE_CLIENT;
		cardconfig &= ~INL_CONFIG_RXANNEX;
		ndev->type = ARPHRD_ETHER;
		clear_bit(FLAG_PROMISC, &priv->flags);
		break;

	case IW_MODE_MASTER:
		bsstype = DOT11_BSSTYPE_INFRA;
		cardmode = INL_MODE_AP;
		cardconfig &= ~INL_CONFIG_RXANNEX;
		ndev->type = ARPHRD_ETHER;
		clear_bit(FLAG_PROMISC, &priv->flags);
		break;

	case IW_MODE_AUTO:
		bsstype = DOT11_BSSTYPE_ANY;
		cardmode = INL_MODE_CLIENT;
		cardconfig &= ~INL_CONFIG_RXANNEX;
		ndev->type = ARPHRD_ETHER;
		clear_bit(FLAG_PROMISC, &priv->flags);
		break;

	case IW_MODE_MONITOR:
		bsstype = DOT11_BSSTYPE_ANY;
		cardmode = INL_MODE_PROMISCUOUS;
		cardconfig |= INL_CONFIG_RXANNEX;
		ndev->type = ARPHRD_IEEE80211;
		set_bit(FLAG_PROMISC, &priv->flags);
		break;

	default:
		return -EOPNOTSUPP;
	}

	/* maybe we can just re-run prism54_mib_init() here? (hvr).
	 * ajfa : except if we want monitor mode.
	 */

	MGT_SET_REQUEST_U32(priv, OID_INL_CONFIG, cardconfig);	

	/* Set the MODE twice, why? */
	MGT_SET_REQUEST_U32(priv, OID_INL_MODE, cardmode);
	MGT_SET_REQUEST_U32(priv, OID_INL_MODE, cardmode);

#warning why isnt the bsstype not set? mcgrof?

	priv->mib.mode = cardmode;
	priv->mib.bsstype = bsstype;

	/* We need to reset the ssid for non-master modes, and also
	   the 802.11d conformance. Except if we are in master or
	   monitor mode */

	if (*uwrq == IW_MODE_MASTER || *uwrq == IW_MODE_MONITOR)
		return 0;

	MGT_SET_REQUEST(priv, DOT11_OID_SSID, &priv->mib.ssid,
			sizeof (struct obj_ssid));
	MGT_SET_REQUEST_U32(priv, OID_INL_DOT11D_CONFORMANCE,
			    priv->mib.conformance);

	return 0;
}

int
prism54_get_mode(struct net_device *ndev, struct iw_request_info *info,
		 __u32 * uwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 bsstype, mode;
	
	/* request the bsstype */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_BSSTYPE, &bsstype);
	
	/* request the mode */
	MGT_GET_REQUEST_U32(priv, OID_INL_MODE, &mode);

	/* translate the cards mode to the wireless mode values as
	   good as possible */

	switch (mode) {
	case INL_MODE_AP:
		*uwrq = IW_MODE_MASTER;
		break;

	case INL_MODE_CLIENT:
		if (bsstype == DOT11_BSSTYPE_IBSS)
			*uwrq = IW_MODE_ADHOC;
		else if (bsstype == DOT11_BSSTYPE_INFRA)
			*uwrq = IW_MODE_INFRA;
		else
			*uwrq = IW_MODE_AUTO;
		break;

	case INL_MODE_PROMISCUOUS:
		*uwrq = IW_MODE_MONITOR;
		break;

	default:		/* DOT11_BSSTYPE_ANY */
		*uwrq = IW_MODE_AUTO;
		break;
	}

	return 0;
}

/* we use DOT11_OID_EDTHRESHOLD. From what I guess the card will not try to
 * emit data if (sensitivity > rssi - noise) (in dBm).
 * prism54_set_sens does not seem to work.
 */

int
prism54_set_sens(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 sens;

	/* by default  the card sets this to 20. */
	sens = vwrq->disabled ? 20 : vwrq->value;

	/* set the ed threshold. */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_EDTHRESHOLD, sens);

	return 0;
}

int
prism54_get_sens(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 sens;

	/* request the ed threshold. */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_EDTHRESHOLD, &sens);

	/* return the data */
	vwrq->value = sens;
	vwrq->disabled = (vwrq->value == 0);
	vwrq->fixed = 1;

	return 0;
}

int
prism54_get_range(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_point *dwrq, char *extra)
{
	struct iw_range *range = (struct iw_range *) extra;
	islpci_private *priv = ndev->priv;
	islpci_qentry_t *entry;
	char *data;
	int i,m;
	struct obj_frequencies *freq;

	/* set the wireless extension version number */
	range->we_version_source = SUPPORTED_WIRELESS_EXT;
	range->we_version_compiled = WIRELESS_EXT;

	/* Now the encoding capabilities */
	range->num_encoding_sizes = 2;
	/* 64(40) bits WEP */
	range->encoding_size[0] = 5;
	/* 128(104) bits WEP */
	range->encoding_size[1] = 13;
	/* 4 keys are allowed */
	range->max_encoding_tokens = 4;

	/* we don't know the quality range... */
	range->max_qual.level = 0;
	range->max_qual.noise = 0;
	range->max_qual.qual = 0;
	/* these value describe an average quality. Needs more tweaking... */
	range->avg_qual.level = -80;	/* -80 dBm */
	range->avg_qual.noise = 0;	/* don't know what to put here */
	range->avg_qual.qual = 0;

	range->sensitivity = 200;

	/* retry limit capabilities */
	range->retry_capa = IW_RETRY_LIMIT | IW_RETRY_LIFETIME;
	range->retry_flags = IW_RETRY_LIMIT;
	range->r_time_flags = IW_RETRY_LIFETIME;

	/* I don't know the range. Put stupid things here */
	range->min_retry = 1;
	range->max_retry = 65535;
	range->min_r_time = 1024;
	range->max_r_time = 65535 * 1024;

	/* Request the device for the supported frequencies
	 * not really revelant since some devices will report the 5 GHz band
	 * frequencies even if they don't support them.
	 */
#define IWMAXFREQ	32
	MGT_GET_REQUEST(priv, DOT11_OID_SUPPORTEDFREQUENCIES, NULL,
			(void **) &freq, (IWMAXFREQ + 1) * sizeof (short) + 2,
			&entry);

	range->num_channels = le16_to_cpu(freq->nr);
	range->num_frequency = le16_to_cpu(freq->nr);

	/* Frequencies are not listed in the right order. The reordering is probably
	 * firmware dependant and thus should be for everyone.
	 */
	m = min(IW_MAX_FREQUENCIES, (int) freq->nr);
	for (i = 0; i < m-12; i++) {
		range->freq[i].m = le16_to_cpu(freq->mhz[12+i]);
		range->freq[i].e = 6;
		range->freq[i].i = i+1;
	}
	for (i = m-12; i < m; i++) {
		range->freq[i].m = le16_to_cpu(freq->mhz[i-m+12]);
		range->freq[i].e = 6;
		range->freq[i].i = i+23;
	}
	
	MGT_FREE_REQUEST(priv, entry);

	/* request the device for the suppoted rates */
	MGT_GET_REQUEST(priv, DOT11_OID_SUPPORTEDRATES, NULL, (void **) &data,
			(IW_MAX_BITRATES + 1) * sizeof (char), &entry);

	/* We got an array of char. It is NULL terminated. */
	i = 0;
	while ((i < IW_MAX_BITRATES) && (*data != 0)) {
		/* the result must be in bps. The card gives us 500Kbps */
		range->bitrate[i] = (__s32) (*data >> 1);
		range->bitrate[i] *= 1000000;
		i++;
		data++;
	}

	range->num_bitrates = i;
	MGT_FREE_REQUEST(priv, entry);

	return 0;
}

/* Set AP address*/

int
prism54_set_wap(struct net_device *ndev, struct iw_request_info *info,
		struct sockaddr *awrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	char bssid[6];

	if (awrq->sa_family != ARPHRD_ETHER)
		return -EINVAL;

	/* prepare the structure for the set object */
	memcpy(&bssid[0], awrq->sa_data, 6);

	/* set the bssid -- does this make sense when in AP mode? */
	MGT_SET_REQUEST(priv, DOT11_OID_BSSID, &bssid, 6);

	return -EINPROGRESS;	/* Call commit handler */
}

/* get AP address*/

int
prism54_get_wap(struct net_device *ndev, struct iw_request_info *info,
		struct sockaddr *awrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	islpci_qentry_t *entry;
	void *data;

	/* request the device for the essid */
	MGT_GET_REQUEST(priv, DOT11_OID_BSSID, NULL, (void **) &data, 6,
			&entry);

	/* return the data */
	memcpy(awrq->sa_data, data, 6);
	awrq->sa_family = ARPHRD_ETHER;

	MGT_FREE_REQUEST(priv, entry);

	return 0;
}

int
prism54_set_scan(struct net_device *dev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	/* hehe the device does this automagicaly */
	return 0;
}

/* a little helper that will translate our data into a card independent
 * format that the Wireless Tools will understand. This was inspired by
 * the "Aironet driver for 4500 and 4800 series cards" (GPL)
 */

inline char *
prism54_translate_bss(struct net_device *ndev, char *current_ev,
		      char *end_buf, struct obj_bss *bss, char noise)
{
	struct iw_event iwe;	/* Temporary buffer */
	short cap;
	
	/* The first entry must be the MAC address */
	memcpy(iwe.u.ap_addr.sa_data, bss->address, 6);
	iwe.u.ap_addr.sa_family = ARPHRD_ETHER;
	iwe.cmd = SIOCGIWAP;
	current_ev =
	    iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_ADDR_LEN);

	/* The following entries will be displayed in the same order we give them */

	/* The ESSID */
	iwe.u.data.length = strnlen(bss->ssid, 32);
	iwe.u.data.flags = 1;
	iwe.cmd = SIOCGIWESSID;
	current_ev = iwe_stream_add_point(current_ev, end_buf, &iwe, bss->ssid);
	
	/* Capabilities */
#define CAP_ESS 0x01
#define CAP_IBSS 0x02
#define CAP_CRYPT 0x03

	/* Mode */
	cap = le16_to_cpu(bss->capinfo);
	iwe.u.mode = 0;
	if(cap & CAP_ESS)
		iwe.u.mode = IW_MODE_MASTER;
	else if(cap & CAP_IBSS)
		iwe.u.mode = IW_MODE_ADHOC;
	iwe.cmd = SIOCGIWMODE;
	if(iwe.u.mode)
		current_ev =
		      iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_UINT_LEN);

	/* Encryption capability */
	if(cap & CAP_CRYPT)
		iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
	else
		iwe.u.data.flags = IW_ENCODE_DISABLED;
	iwe.u.data.length = 0;
	iwe.cmd = SIOCGIWENCODE;
	current_ev = iwe_stream_add_point(current_ev, end_buf, &iwe, NULL);

	/* Add frequency. (short) bss->channel is the frequency in MHz */
	iwe.u.freq.m = bss->channel;
	iwe.u.freq.e = 6;
	iwe.cmd = SIOCGIWFREQ;
	current_ev =
	    iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_FREQ_LEN);

	/* Add quality statistics */
	iwe.u.qual.level = bss->rssi;
	iwe.u.qual.noise = noise;
	/* do a simple SNR for quality */
	iwe.u.qual.qual = bss->rssi - noise;
	iwe.cmd = IWEVQUAL;
	current_ev =
	    iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_QUAL_LEN);

	return current_ev;
}

int
prism54_get_scan(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	islpci_qentry_t *entry;
	int i;
	struct obj_bsslist *bsslist;
	u32 noise = 0;
	char *current_ev = extra;

	/* first get the noise value. We will use it to report the link quality */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_NOISEFLOOR, &noise);
	
	/* Ask the device for a list of known bss. We can report at most
	 * IW_MAX_AP=64 to the range struct. But the device won't repport anything
	 * if you change the value of MAXBSS=24. Anyway 24 AP It is probably enough.
	 */
#define IWMAXBSS	24

	MGT_GET_REQUEST(priv, DOT11_OID_BSSLIST, NULL, (void **) &bsslist,
			8 + 4 + IWMAXBSS * sizeof (struct obj_bss), &entry);

	/* ok now, scan the list and translate its info */
	for (i = 0; i < min(IW_MAX_AP, (int) le32_to_cpu(bsslist->nr)); i++)
		current_ev = prism54_translate_bss(ndev, current_ev,
						   extra + IW_SCAN_MAX_DATA,
						   &(bsslist->bsslist[i]),
						   noise);
	MGT_FREE_REQUEST(priv, entry);
	dwrq->length = (current_ev - extra);
	dwrq->flags = 0;	/* todo */

	return 0;
}

int
prism54_set_essid(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct obj_ssid essid_obj;

	memset(essid_obj.octets, 0, 33);

	/* Check if we were asked for `any' */
	if (dwrq->flags) {
		/* we must set the SSID. */
		if (dwrq->length > min(33, IW_ESSID_MAX_SIZE + 1))
			return -E2BIG;
		essid_obj.length = dwrq->length - 1;
		memcpy(essid_obj.octets, extra, essid_obj.length);
	} else {
		/* we must clear the ssid */
		essid_obj.length = 0;
	}

	/* request the device to set the essid */
	MGT_SET_REQUEST(priv, DOT11_OID_SSID, &essid_obj,
			sizeof (struct obj_ssid));

	/* now that it succeeded we can update our driver state */
	memcpy(&priv->mib.ssid, &essid_obj, sizeof (struct obj_ssid));

	return 0;
}

int
prism54_get_essid(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	islpci_qentry_t *entry;
	struct obj_ssid *essid_obj;

	/* request the device for the essid */
	MGT_GET_REQUEST(priv, DOT11_OID_SSID, NULL,
			(void **) &essid_obj, sizeof (struct obj_ssid), &entry);

	/* copy the data to the return structure */
	dwrq->flags = 1;	/* set ESSID to ON for Wireless Extensions */
	dwrq->length = essid_obj->length + 1;

	memcpy(extra, essid_obj->octets, dwrq->length);

	MGT_FREE_REQUEST(priv, entry);

	return 0;
}

/* Set the allowed Bitrates - BROKEN - disabled */

int
prism54_set_rate(struct net_device *ndev,
		 struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{

	islpci_private *priv = ndev->priv;

	if (vwrq->value == -1) {
		/* auto mode. No limit. */

	}

	/* Now, check if we want a fixed or auto value */
	if (vwrq->fixed == 0) {
		/* set the maximum allowed bitrate */

	} else {
		/* Fixed rate  mode */

	}

	/* we have a frequency value in bps. The card wants 500Kbps */
	/* rate = (u8) (vwrq->value/500000); */

	MGT_SET_REQUEST_U32(priv, DOT11_OID_ALOFT_FIXEDRATE, vwrq->value);

	return 0;
}

/* Get the current bit rate */

int
prism54_get_rate(struct net_device *ndev,
		 struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;

	/* Get the current bit rate */
	MGT_GET_REQUEST_U32(priv, GEN_OID_LINKSTATE, &vwrq->value);
	vwrq->value *= 500000;
	
	/* request the device for the fixed rate */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_ALOFT_FIXEDRATE, &u);

	/* if *data = -1 we are not in fixed rate mode. Report it. */
	vwrq->fixed = (u != -1);
	
	return 0;
}

int
prism54_set_rts(struct net_device *ndev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	/* set the rts threshold */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_RTSTHRESH, vwrq->value);

	return 0;
}

int
prism54_get_rts(struct net_device *ndev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;

	/* get the rts threshold */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_RTSTHRESH, &u);
	vwrq->value = u;
	
	return 0;
}

int
prism54_set_frag(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	/*  set the fragment threshold */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_RTSTHRESH, vwrq->value);

	return 0;
}

int
prism54_get_frag(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;

	/* get the fragment threshold */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_FRAGTHRESH, &u);
	vwrq->value = u;
	
	return 0;
}

/* Here we have (min,max) = max retries for (small frames, big frames). Where
 * big frame <=>  bigger than the rts threshold
 * small frame <=>  smaller than the rts threshold
 * This is not really the behavior expected by the wireless tool but it seems
 * to be a common behavior in other drivers.
 * 
 * It seems that playing with this tends to hang the card -> DISABLED
 */

int
prism54_set_retry(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	int slimit = 0, llimit = 0;	/* short and long limit */
	int lifetime = 0;

	if (vwrq->disabled)
		/* we cannot disable this feature */
		return -EINVAL;

	if (vwrq->flags & IW_RETRY_LIMIT) {
		if (vwrq->flags & IW_RETRY_MIN)
			slimit = vwrq->value;
		else if (vwrq->flags & IW_RETRY_MAX)
			llimit = vwrq->value;
		else {
			/* we are asked to set both */
			slimit = vwrq->value;
			llimit = vwrq->value;
		}
	}
	if (vwrq->flags & IW_RETRY_LIFETIME)
		/* Wireless tools use us unit while the device uses 1024 us unit */
		lifetime = vwrq->value / 1024;

	/* now set what is requested */

	if (slimit != 0)
		MGT_SET_REQUEST_U32(priv, DOT11_OID_SHORTRETRIES, slimit);
	if (llimit != 0)
		MGT_SET_REQUEST_U32(priv, DOT11_OID_LONGRETRIES, llimit);
	if (lifetime != 0)
		MGT_SET_REQUEST_U32(priv, DOT11_OID_MAXTXLIFETIME, lifetime);

	return 0;
}

int
prism54_get_retry(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;
	unsigned long oid;

	vwrq->disabled = 0;	/* It cannot be disabled */

	if ((vwrq->flags & IW_RETRY_TYPE) == IW_RETRY_LIFETIME)
		/* we are asked for the life time */
		oid = DOT11_OID_MAXTXLIFETIME;
	else if ((vwrq->flags & IW_RETRY_MAX))
		/* we are asked for the long retry limit */
		oid = DOT11_OID_LONGRETRIES;
	else
		/* default. get the  short retry limit */
		oid = DOT11_OID_SHORTRETRIES;

	/* now get the selected oid */
	MGT_GET_REQUEST_U32(priv, oid, &u);

	/* return the result */
	switch (oid) {
	case DOT11_OID_MAXTXLIFETIME:
		/* device uses the 1024 us unit */
		vwrq->value = u * 1024;
		vwrq->flags = IW_RETRY_LIFETIME;
		break;

	case DOT11_OID_LONGRETRIES:
		vwrq->value = u;
		vwrq->flags = IW_RETRY_LIMIT | IW_RETRY_MAX;
		break;

	default:
		vwrq->value = u;
		vwrq->flags = IW_RETRY_LIMIT | IW_RETRY_MIN;
	}

	return 0;
}

/* Now the WEP functions. We begin with some helpers. */

int
getdefkeyid(islpci_private *priv, int *res)
{
	u32 u;

	/* get the default key id */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_DEFKEYID, &u);
	*res = u;
	
	return 0;
}

int
setdefkeyid(islpci_private *priv, int arg)
{
	/* set the default key id */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_DEFKEYID, arg);

	return 0;
}

int
getdefkeyx(islpci_private *priv, struct obj_key *res, int index)
{
	struct obj_key *pkey;
	unsigned long oid;
	islpci_qentry_t *entry;
	
	switch (index) {
	case 0:
		oid = DOT11_OID_DEFKEY1;
		break;

	case 1:
		oid = DOT11_OID_DEFKEY2;
		break;

	case 2:
		oid = DOT11_OID_DEFKEY3;
		break;

	case 3:
		oid = DOT11_OID_DEFKEY4;
		break;

	default:
		return -EOPNOTSUPP;
	}

	/* get the corresponding key */
	MGT_GET_REQUEST(priv, oid, NULL, (void **) &pkey,
			sizeof (struct obj_key), &entry);
	memcpy(res, pkey, sizeof (struct obj_key));
	MGT_FREE_REQUEST(priv, entry);
	return 0;
}

int
setdefkeyx(islpci_private *priv, struct obj_key *key, int index)
{
	unsigned long oid;
	
	switch (index) {
	case 0:
		oid = DOT11_OID_DEFKEY1;
		break;

	case 1:
		oid = DOT11_OID_DEFKEY2;
		break;

	case 2:
		oid = DOT11_OID_DEFKEY3;
		break;

	case 3:
		oid = DOT11_OID_DEFKEY4;
		break;

	default:
		return -EOPNOTSUPP;
	}

	/* set the corresponding key */
	MGT_SET_REQUEST(priv, oid, key, sizeof (struct obj_key));
	memcpy(&(priv->mib.defkey[index]), key, sizeof(struct obj_key));
	
	return 0;
}

int
getauthenten(islpci_private *priv, int *res)
{
	u32 u;

	/* get the auth method */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_AUTHENABLE, &u);
	*res = u;

	return 0;
}

int
setauthenten(islpci_private *priv, int arg)
{
	/* set the auth method */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_AUTHENABLE, arg);
	priv->mib.authenable = arg;
	
	return 0;
}

int
getprivfilter(islpci_private *priv, int *res)
{
	u32 u;
	
	/* get the privacy filter */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_EXUNENCRYPTED,&u);
	*res = u;
	
	return 0;
}

int
setprivfilter(islpci_private *priv, int arg)
{
	/* set the privacy filter */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_EXUNENCRYPTED, arg);
	priv->mib.exunencrypted = arg;

	return 0;
}

int
getprivinvoke(islpci_private *priv, int *res)
{
	u32 u;

	/* get the privacy_invoked value */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_PRIVACYINVOKED, &u);
	*res = u;
	
	return 0;
}

int
setprivinvoke(islpci_private *priv, int arg)
{
	/* set the privacy invoked value */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_PRIVACYINVOKED, arg);
	priv->mib.privacyinvoked = arg;

	return 0;
}

/* And now the real functions */

int
prism54_set_encode(struct net_device *ndev, struct iw_request_info *info,
		   struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	int rvalue;
	int force = 0;

	int authen = 0;
	int invoke = 0;
	int exunencrypt = 0;

	/* with the new API, it's impossible to get a NULL pointer.
	 * New version of iwconfig set the IW_ENCODE_NOKEY flag
	 * when no key is given, but older versions don't. */

	if (dwrq->length > 0) {
		/* we have a key to set */
		int index = (dwrq->flags & IW_ENCODE_INDEX) - 1;
		int current_index;
		struct obj_key key = { DOT11_PRIV_WEP, 0, "" };

		/* get the current key index */
		if (rvalue = getdefkeyid(priv, &current_index), rvalue != 0)
			return rvalue;
		/* Verify that the key is not marked as invalid */
		if (!(dwrq->flags & IW_ENCODE_NOKEY)) {
			memcpy(key.key, extra, dwrq->length);
			key.length = dwrq->length > sizeof (key.key) ?
			    sizeof (key.key) : dwrq->length;

			if ((index < 0) || (index > 3))
				/* no index provided use the current one */
				index = current_index;

			/* now send the key to the card  */
			if (rvalue = setdefkeyx(priv, &key, index), rvalue != 0)
				return rvalue;
		}
		/*
		 * If a valid key is set, encryption should be enabled 
		 * (user may turn it off later).
		 * This is also how "iwconfig ethX key on" works
		 */
		if ((index == current_index) && (key.length > 0))
			force = 1;
	} else {
		int index = (dwrq->flags & IW_ENCODE_INDEX) - 1;
		if ((index >= 0) && (index <= 3)) {
			/* we want to set the key index */
			if (rvalue = setdefkeyid(priv, index), rvalue != 0)
				return rvalue;
		} else {
			if (!dwrq->flags & IW_ENCODE_MODE) {
				/* we cannot do anything. Complain. */
				return -EINVAL;
			}
		}
	}

	/* now read the flags     */
	if (dwrq->flags & IW_ENCODE_DISABLED) {
		/* Encoding disabled, 
		 * authen = 0;
		 * invoke = 0;
		 * exunencrypt = 0; */
	}
	if ((dwrq->flags & IW_ENCODE_OPEN) || force)
		/* Encode but accept non-encoded packets. No auth */
		invoke = 1;
	if ((dwrq->flags & IW_ENCODE_RESTRICTED) || force) {
		/* Refuse non-encoded packets. Auth */
		authen = DOT11_AUTH_BOTH;
		invoke = 1;
		exunencrypt = 1;
	}
	/* do the change if requested  */
	if ((dwrq->flags & IW_ENCODE_MODE) || force) {
		if (rvalue = setauthenten(priv, authen), rvalue != 0)
			return rvalue;
		if (rvalue = setprivinvoke(priv, invoke), rvalue != 0)
			return rvalue;
		if (rvalue = setprivfilter(priv, exunencrypt), rvalue != 0)
			return rvalue;
	}
	return 0;
}

int
prism54_get_encode(struct net_device *ndev, struct iw_request_info *info,
		   struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	int rvalue;
	struct obj_key key;

	int devindex;
	int index = (dwrq->flags & IW_ENCODE_INDEX) - 1;
	int authen = 0;
	int invoke = 0;
	int exunencrypt = 0;

	/* first get the flags */

	if (rvalue = getauthenten(priv, &authen), rvalue != 0)
		return rvalue;
	if (rvalue = getprivinvoke(priv, &invoke), rvalue != 0)
		return rvalue;
	if (rvalue = getprivfilter(priv, &exunencrypt), rvalue != 0)
		return rvalue;

	/* the only interresting value for authen is DOT11_AUTH_BOTH */
	if (authen == DOT11_AUTH_BOTH)
		authen = 1;
	else
		authen = 0;

	if (invoke && authen && exunencrypt)
		dwrq->flags = IW_ENCODE_RESTRICTED;
	else if (!authen && !exunencrypt) {
		if (invoke)
			dwrq->flags = IW_ENCODE_OPEN;
		else
			dwrq->flags = IW_ENCODE_DISABLED;
	} else
		/* The card should not work in this state */
		dwrq->flags = 0;

	/* get the current device key index */
	if (rvalue = getdefkeyid(priv, &devindex), rvalue != 0)
		return rvalue;

	/* Now get the key, return it */
	if ((index < 0) || (index > 3))
		/* no index provided, use the current one */
		index = devindex;
	if (!getdefkeyx(priv, &key, index)) {
		dwrq->length = key.length;
		memcpy(extra, key.key, dwrq->length);
	} else
		dwrq->length = 0;
	/* return the used key index */
	dwrq->flags |= devindex + 1;
	return 0;
}


int prism54_reset(struct net_device *ndev, struct iw_request_info *info,
		      struct iwreq *wrq, char *extra)
{
	islpci_reset(ndev->priv);

	return 0;
}

int prism54_set_beacon(struct net_device *ndev, struct iw_request_info *info,
		        struct iwreq *wrq, char *extra)
{
	MGT_SET_REQUEST_U32(ndev->priv, DOT11_OID_BEACONPERIOD, (u32)*extra);
	
	return 0;
}

int prism54_get_beacon(struct net_device *ndev, struct iw_request_info *info,
		      struct iwreq *wrq, char *extra)
{
	MGT_GET_REQUEST_U32(ndev->priv, DOT11_OID_BEACONPERIOD, (u32*)extra);
	
	return 0;
}

int prism54_set_psm(struct net_device *ndev, struct iw_request_info *info,
		      struct iwreq *wrq, char *extra)
{
	MGT_SET_REQUEST_U32(ndev->priv, DOT11_OID_PSM, (u32)*extra);
	
	return 0;
}

int prism54_get_psm(struct net_device *ndev, struct iw_request_info *info,
		      struct iwreq *wrq, char *extra)
{
	MGT_GET_REQUEST_U32(ndev->priv, DOT11_OID_PSM, (u32*)extra);
	
	return 0;
}


/* Translate a TRAP oid into a wireless event. Called in islpci_mgt_receive. */

#define STR_ADD(dest,src,length)do { \
	memcpy(&(dest)[(length)], (src), sizeof((src))-1);\
	(length) += sizeof((src))-1;\
	} while (0)

#define STR_ADD_ADDRESS(dest,addr,length)do { \
	int _i;	\
	for ( _i = 0; _i < 5; _i++ )	\
		sprintf(&(dest)[(length)+3*_i],"%2.2X:",(unsigned char)((addr)[_i]));\
	sprintf(&(dest)[(length)+15],"%2.2X", (unsigned char)((addr)[5]));	\
	(length) += 17;	\
	} while (0)

#define FORMAT_EVENT(dest,str,mlme,length,error)do { \
	STR_ADD((dest), (str), (length)); \
	if (priv->mib.mode == INL_MODE_CLIENT)	\
		STR_ADD((dest)," to ", (length));	\
	else	\
		STR_ADD((dest), " from ", (length));	\
	STR_ADD_ADDRESS((dest), (mlme)->address, (length));	\
	if ((error)){	\
		if ((mlme)->code)	\
			STR_ADD((dest), " : FAILED ", (length));	\
		else	\
			STR_ADD((dest), " : SUCCES ", (length));	\
	}	\
	} while (0)

#define	SEND_FORMATED_EVENT(dev, str, mlme,error) do{	\
	union iwreq_data _wrqu;	\
	_wrqu.data.pointer = kmalloc(IW_CUSTOM_MAX, GFP_ATOMIC);	\
	_wrqu.data.length = 0;	\
	FORMAT_EVENT(_wrqu.data.pointer, (str), (mlme),	\
	       _wrqu.data.length,(error));	\
	wireless_send_event((dev), IWEVCUSTOM, &_wrqu, _wrqu.data.pointer);	\
	} while (0)

#define	SEND_SIMPLE_EVENT(dev, str) do{	\
	union iwreq_data _wrqu;	\
	_wrqu.data.pointer = kmalloc(IW_CUSTOM_MAX, GFP_ATOMIC);	\
	_wrqu.data.length = 0;	\
	STR_ADD(_wrqu.data.pointer, (str),_wrqu.data.length);\
	wireless_send_event((dev), IWEVCUSTOM, &_wrqu, _wrqu.data.pointer);	\
	} while (0)

void
prism54_send_trap(islpci_private *priv, u32  oid, char *data, u32 length)
{
	struct net_device *ndev = priv->ndev;
	
	/* I think all trapable objects are listed here. Not all are exploited.
	 * Some oids have a EX version. The difference is that they are emitted
	 * in DOT11_MLME_EXTENDED mode (set with DOT11_OID_MLMEAUTOLEVEL)
	 * with more info.
	 * The few events already defined by the wireless tools are not really
	 * suited. We use the more flexible custom event facility.
	 */
	
	switch (oid) {
	
	case GEN_OID_LINKSTATE:
		/* The bitrate has changed from O to something or from something to 0.
		 * Means that the connection has been established or lost.
		 * To report it through the SIOCGIWAP event we would need to retrieve
		 * AP's MAC. But I don't think it would be a good idea in interrupt
		 * context. Thus we report it with through a cutom event.
		 */
		if ((double) *data)
			SEND_SIMPLE_EVENT(ndev, "Link established");	
		else
			SEND_SIMPLE_EVENT(ndev, "Link lost");
		break;

	case DOT11_OID_MICFAILURE:
		SEND_SIMPLE_EVENT(ndev, "Mic failure");
		break;
	
	case DOT11_OID_DEAUTHENTICATE : 
		SEND_FORMATED_EVENT(ndev, "DeAuthenticate request",
		         (struct obj_mlme*)data, 0);
		break;
	
	case DOT11_OID_AUTHENTICATE :
		SEND_FORMATED_EVENT(ndev, "Authenticate request",
		         (struct obj_mlme*)data, 1);
		break;
	
	case DOT11_OID_DISASSOCIATE :
		SEND_FORMATED_EVENT(ndev, "Disassociate request",
		         (struct obj_mlme*)data, 0);
		break;
	
	case DOT11_OID_ASSOCIATE :
		SEND_FORMATED_EVENT(ndev, "Associate request",
		          (struct obj_mlme*)data, 1);
		break;
		
	case DOT11_OID_REASSOCIATE :
		SEND_FORMATED_EVENT(ndev, "ReAssociate request",
		          (struct obj_mlme*)data, 1);
		break;
		
	case DOT11_OID_BEACON :
		SEND_FORMATED_EVENT(ndev, "Received a beacon from an unkown AP",
			(struct obj_mlme*)data, 0);
		break;
		
	case DOT11_OID_PROBE :
		/* we received a probe from a client. */
		SEND_FORMATED_EVENT(ndev, "Received a beacon from an unkown AP",
			(struct obj_mlme*)data, 0);
		break;
		
	case DOT11_OID_DEAUTHENTICATEEX :
		SEND_FORMATED_EVENT(ndev, "DeAuthenticate request",
		         (struct obj_mlmeex*)data, 0);
		break;
		
	case DOT11_OID_AUTHENTICATEEX :
		SEND_FORMATED_EVENT(ndev, "Authenticate request",
		         (struct obj_mlmeex*)data, 1);
		break;
		
	case DOT11_OID_DISASSOCIATEEX :
		SEND_FORMATED_EVENT(ndev, "Disassociate request",
		         (struct obj_mlmeex*)data, 0);
		break;
		
	case DOT11_OID_ASSOCIATEEX :
		SEND_FORMATED_EVENT(ndev, "Associate request",
		         (struct obj_mlmeex*)data, 1);
		break;
	
	case DOT11_OID_REASSOCIATEEX :
		SEND_FORMATED_EVENT(ndev, "Reassociate request",
		         (struct obj_mlmeex*)data, 1);
		break;
		
	default:
		return;
	}
	
		
}

int
prism54_ioctl(struct net_device *ndev, struct ifreq *rq, int cmd)
{
	/* should we really support this old shit ? */
	return -EOPNOTSUPP;
}
