/*  $Header: /var/lib/cvs/prism54-ng/ksrc/isl_ioctl.h,v 1.15 2003/11/10 07:56:05 ajfa Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *            (C) 2003 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _ISL_IOCTL_H
#define _ISL_IOCTL_H

#include <linux/wireless.h>
#if WIRELESS_EXT > 12
#include <net/iw_handler.h>	// New driver API
#endif				/* WIRELESS_EXT > 12 */

#include "islpci_mgt.h"
#include "bloboid.h"		/* additional types and defs for 38xx firmware */

#define SUPPORTED_WIRELESS_EXT                  16

void prism54_mib_init_defaults(struct islpci_mib *mib);
int prism54_mib_init(islpci_private *);

struct iw_statistics *prism54_get_wireless_stats(struct net_device *);

int prism54_commit(struct net_device *, struct iw_request_info *, char *,
		   char *);
int prism54_get_name(struct net_device *, struct iw_request_info *, char *,
		     char *);
int prism54_set_freq(struct net_device *, struct iw_request_info *,
		     struct iw_freq *, char *);
int prism54_get_freq(struct net_device *, struct iw_request_info *,
		     struct iw_freq *, char *);
int prism54_set_mode(struct net_device *, struct iw_request_info *,
		     __u32 *, char *);
int prism54_get_mode(struct net_device *, struct iw_request_info *,
		     __u32 * u, char *);
int prism54_set_sens(struct net_device *, struct iw_request_info *,
		     struct iw_param *, char *);
int prism54_get_sens(struct net_device *, struct iw_request_info *,
		     struct iw_param *, char *);
int prism54_get_range(struct net_device *, struct iw_request_info *,
		      struct iw_point *, char *);
int prism54_set_wap(struct net_device *, struct iw_request_info *,
		    struct sockaddr *, char *);
int prism54_get_wap(struct net_device *, struct iw_request_info *,
		    struct sockaddr *, char *);
int prism54_set_scan(struct net_device *, struct iw_request_info *,
		     struct iw_param *, char *);
int prism54_get_scan(struct net_device *, struct iw_request_info *,
		     struct iw_point *, char *);
int prism54_set_essid(struct net_device *, struct iw_request_info *,
		      struct iw_point *, char *);
int prism54_get_essid(struct net_device *, struct iw_request_info *,
		      struct iw_point *, char *);
int prism54_set_rate(struct net_device *, struct iw_request_info *,
		     struct iw_param *, char *);
int prism54_get_rate(struct net_device *, struct iw_request_info *,
		     struct iw_param *, char *);
int prism54_set_rts(struct net_device *, struct iw_request_info *,
		    struct iw_param *, char *);
int prism54_get_rts(struct net_device *, struct iw_request_info *,
		    struct iw_param *, char *);
int prism54_set_frag(struct net_device *, struct iw_request_info *,
		     struct iw_param *, char *);
int prism54_get_frag(struct net_device *, struct iw_request_info *,
		     struct iw_param *, char *);
int prism54_set_retry(struct net_device *, struct iw_request_info *,
		      struct iw_param *, char *);
int prism54_get_retry(struct net_device *, struct iw_request_info *,
		      struct iw_param *, char *);
int prism54_set_encode(struct net_device *, struct iw_request_info *,
		       struct iw_point *, char *);
int prism54_get_encode(struct net_device *, struct iw_request_info *,
		       struct iw_point *, char *);

int prism54_reset(struct net_device *, struct iw_request_info *,
		      struct iwreq *, char *);
int prism54_set_beacon(struct net_device *, struct iw_request_info *,
		      struct iwreq *, char *);
int prism54_get_beacon(struct net_device *, struct iw_request_info *,
		      struct iwreq *, char *);
int prism54_set_psm(struct net_device *, struct iw_request_info *,
		      struct iwreq *, char *);
int prism54_get_psm(struct net_device *, struct iw_request_info *,
		      struct iwreq *, char *);

int prism54_ioctl(struct net_device *, struct ifreq *, int);

void prism54_send_trap(islpci_private *, u32, char *, u32);

#if WIRELESS_EXT > 12
static const iw_handler prism54_handler[] = {
	(iw_handler) prism54_commit,	/* SIOCSIWCOMMIT */
	(iw_handler) prism54_get_name,	/* SIOCGIWNAME */
	(iw_handler) NULL,	/* SIOCSIWNWID */
	(iw_handler) NULL,	/* SIOCGIWNWID */
	(iw_handler) prism54_set_freq,	/* SIOCSIWFREQ */
	(iw_handler) prism54_get_freq,	/* SIOCGIWFREQ */
	(iw_handler) prism54_set_mode,	/* SIOCSIWMODE */
	(iw_handler) prism54_get_mode,	/* SIOCGIWMODE */
	(iw_handler) prism54_set_sens,	/* SIOCSIWSENS */
	(iw_handler) prism54_get_sens,	/* SIOCGIWSENS */
	(iw_handler) NULL,	/* SIOCSIWRANGE */
	(iw_handler) prism54_get_range,	/* SIOCGIWRANGE */
	(iw_handler) NULL,	/* SIOCSIWPRIV */
	(iw_handler) NULL,	/* SIOCGIWPRIV */
	(iw_handler) NULL,	/* SIOCSIWSTATS */
	(iw_handler) NULL,	/* SIOCGIWSTATS */
#if WIRELESS_EXT > 15
	iw_handler_set_spy,	/* SIOCSIWSPY */
	iw_handler_get_spy,	/* SIOCGIWSPY */
	iw_handler_set_thrspy,	/* SIOCSIWTHRSPY */
	iw_handler_get_thrspy,	/* SIOCGIWTHRSPY */
#else				/* WIRELESS_EXT > 15 */
	(iw_handler) NULL,	/* SIOCSIWSPY */
	(iw_handler) NULL,	/* SIOCGIWSPY */
	(iw_handler) NULL,	/* -- hole -- */
	(iw_handler) NULL,	/* -- hole -- */
#endif				/* WIRELESS_EXT > 15 */
	(iw_handler) prism54_set_wap,	/* SIOCSIWAP */
	(iw_handler) prism54_get_wap,	/* SIOCGIWAP */
	(iw_handler) NULL,	/* -- hole -- */
	(iw_handler) NULL,	/* SIOCGIWAPLIST depreciated */
#if WIRELESS_EXT > 13
	(iw_handler) prism54_set_scan,	/* SIOCSIWSCAN */
	(iw_handler) prism54_get_scan,	/* SIOCGIWSCAN */
#else				/* WIRELESS_EXT > 13 */
	(iw_handler) NULL,	/* SIOCSIWSCAN */
	(iw_handler) NULL,	/* SIOCGIWSCAN */
#endif				/* WIRELESS_EXT > 13 */
	(iw_handler) prism54_set_essid,	/* SIOCSIWESSID */
	(iw_handler) prism54_get_essid,	/* SIOCGIWESSID */
	(iw_handler) NULL,	/* SIOCSIWNICKN */
	(iw_handler) NULL,	/* SIOCGIWNICKN */
	(iw_handler) NULL,	/* -- hole -- */
	(iw_handler) NULL,	/* -- hole -- */
	(iw_handler) NULL,	/* SIOCSIWRATE */
	(iw_handler) prism54_get_rate,	/* SIOCGIWRATE */
	(iw_handler) prism54_set_rts,	/* SIOCSIWRTS */
	(iw_handler) prism54_get_rts,	/* SIOCGIWRTS */
	(iw_handler) prism54_set_frag,	/* SIOCSIWFRAG */
	(iw_handler) prism54_get_frag,	/* SIOCGIWFRAG */
	(iw_handler) NULL,	/* SIOCSIWTXPOW */
	(iw_handler) NULL,	/* SIOCGIWTXPOW */
	(iw_handler) prism54_set_retry,	/* SIOCSIWRETRY */
	(iw_handler) prism54_get_retry,	/* SIOCGIWRETRY */
	(iw_handler) prism54_set_encode,	/* SIOCSIWENCODE */
	(iw_handler) prism54_get_encode,	/* SIOCGIWENCODE */
	(iw_handler) NULL,	/* SIOCSIWPOWER */
	(iw_handler) NULL,	/* SIOCGIWPOWER */
};

/* The low order bit identify a SET (0) or a GET (1) ioctl.  */

#define PRISM54_RESET		SIOCIWFIRSTPRIV
#define PRISM54_GET_BEACON	SIOCIWFIRSTPRIV+1
#define PRISM54_SET_BEACON	SIOCIWFIRSTPRIV+2
#define PRISM54_GET_PSM		SIOCIWFIRSTPRIV+3
#define PRISM54_SET_PSM		SIOCIWFIRSTPRIV+4

static const struct iw_priv_args prism54_private_args[] = {
/*{ cmd, set_args, get_args, name } */
	{ PRISM54_RESET, 0, 0, "reset" },
	{ PRISM54_GET_BEACON, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	 "getBeaconPeriod" },
	{ PRISM54_SET_BEACON, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0,
	 "setBeaconPeriod" },
	{ PRISM54_GET_PSM, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	 "getPowerMode" },
	{ PRISM54_SET_PSM, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0,
	 "setPowerMode" },
}; 

static const iw_handler	prism54_private_handler[] =
{
	(iw_handler) prism54_reset,
	(iw_handler) prism54_get_beacon,
	(iw_handler) prism54_set_beacon,
	(iw_handler) prism54_get_psm,
	(iw_handler) prism54_set_psm,
};

static const struct iw_handler_def prism54_handler_def = {
	.num_standard = sizeof (prism54_handler) / sizeof (iw_handler),
	.num_private = sizeof(prism54_private_handler)/sizeof(iw_handler),
	.num_private_args = sizeof(prism54_private_args)/sizeof(struct iw_priv_args),
	.standard = (iw_handler *) prism54_handler,
	.private = (iw_handler *) prism54_private_handler,
	.private_args = (struct iw_priv_args *) prism54_private_args,
};



#endif				/* WIRELESS_EXT > 12 */

#endif				/* _ISL_IOCTL_H */
