/*  $Header: /var/lib/cvs/prism54-ng/ksrc/islpci_dev.c,v 1.27 2003/11/14 03:48:28 hvr Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *  Copyright (C) 2003 Herbert Valerio Riedel <hvr@gnu.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/version.h>
#include <linux/module.h>

#include <linux/netdevice.h>
#include <linux/pci.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>
#include <linux/if_arp.h>

#include <asm/io.h>

#include "isl_gen.h"
#include "isl_38xx.h"
#include "isl_ioctl.h"
#include "islpci_dev.h"
#include "islpci_mgt.h"
#include "islpci_eth.h"

#ifdef WDS_LINKS
#include <isl_mgt.h>
#include <isl_wds.h>
#endif

#define ISL3877_IMAGE_FILE	"isl3877"
#define ISL3890_IMAGE_FILE	"isl3890"

/******************************************************************************
    Device Interrupt Handler
******************************************************************************/

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,5,75))
irqreturn_t
#else
# define IRQ_HANDLED
# define IRQ_NONE
void
#endif
islpci_interrupt(int irq, void *config, struct pt_regs *regs)
{
	u32 reg;
	islpci_private *priv = config;
	void *device = priv->device_base;

	/* received an interrupt request on a shared IRQ line
	 * first check whether the device is in sleep mode */
	if (reg = readl(device + ISL38XX_CTRL_STAT_REG),
	    reg & ISL38XX_CTRL_STAT_SLEEPMODE)
		/* device is in sleep mode, IRQ was generated by someone else */
		return IRQ_NONE;

	/* lock the interrupt handler */
	spin_lock(&priv->slock);

	/* check whether there is any source of interrupt on the device */
	reg = readl(device + ISL38XX_INT_IDENT_REG);

	/* also check the contents of the Interrupt Enable Register, because this
	 * will filter out interrupt sources from other devices on the same irq ! */
	reg &= readl(device + ISL38XX_INT_EN_REG);

	if (reg &= ISL38XX_INT_SOURCES, reg != 0) {
		/* reset the request bits in the Identification register */
		isl38xx_w32_flush(device, reg, ISL38XX_INT_ACK_REG);

#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_FUNCTION_CALLS,
		      "IRQ: Identification register 0x%p 0x%x \n", device, reg);
#endif

		/* check for each bit in the register separately */
		if (reg & ISL38XX_INT_IDENT_UPDATE) {
#if VERBOSE > SHOW_ERROR_MESSAGES
			/* Queue has been updated */
			DEBUG(SHOW_TRACING, "IRQ: Update flag \n");

			DEBUG(SHOW_QUEUE_INDEXES,
			      "CB drv Qs: [%i][%i][%i][%i][%i][%i]\n",
			      le32_to_cpu(priv->control_block->
					  driver_curr_frag[0]),
			      le32_to_cpu(priv->control_block->
					  driver_curr_frag[1]),
			      le32_to_cpu(priv->control_block->
					  driver_curr_frag[2]),
			      le32_to_cpu(priv->control_block->
					  driver_curr_frag[3]),
			      le32_to_cpu(priv->control_block->
					  driver_curr_frag[4]),
			      le32_to_cpu(priv->control_block->
					  driver_curr_frag[5])
			    );

			DEBUG(SHOW_QUEUE_INDEXES,
			      "CB dev Qs: [%i][%i][%i][%i][%i][%i]\n",
			      le32_to_cpu(priv->control_block->
					  device_curr_frag[0]),
			      le32_to_cpu(priv->control_block->
					  device_curr_frag[1]),
			      le32_to_cpu(priv->control_block->
					  device_curr_frag[2]),
			      le32_to_cpu(priv->control_block->
					  device_curr_frag[3]),
			      le32_to_cpu(priv->control_block->
					  device_curr_frag[4]),
			      le32_to_cpu(priv->control_block->
					  device_curr_frag[5])
			    );
#endif
			
			/* cleanup the data low transmit queue */
	 		islpci_eth_cleanup_transmit(priv, priv->control_block);
			
			/* device is in active state, update the powerstate flag */
			priv->powerstate = ISL38XX_PSM_ACTIVE_STATE;

			/* check all three queues in priority order
			 * call the PIMFOR receive function until the
			 * queue is empty */
			while (isl38xx_in_queue(priv->control_block,
						ISL38XX_CB_RX_MGMTQ) != 0) {
#if VERBOSE > SHOW_ERROR_MESSAGES
				DEBUG(SHOW_TRACING,
				      "Received frame in Management Queue\n");
#endif
				islpci_mgt_receive(priv);

				/* call the pimfor transmit function
				 * for processing the next management
				 * frame in the shadow queue */
				islpci_mgt_transmit(priv);
			}

			while (isl38xx_in_queue(priv->control_block,
						ISL38XX_CB_RX_DATA_LQ) != 0) {
#if VERBOSE > SHOW_ERROR_MESSAGES
				DEBUG(SHOW_TRACING,
				      "Received frame in Data Low Queue \n");
#endif
				islpci_eth_receive(priv);
			}

			/* check whether the data transmit queues were full */
			if (priv->data_low_tx_full) {
				/* check whether the transmit is not full anymore */
				if (ISL38XX_CB_TX_QSIZE -
				    isl38xx_in_queue(priv->control_block,
						     ISL38XX_CB_TX_DATA_LQ) >=
				    ISL38XX_MIN_QTHRESHOLD) {
					/* nope, the driver is ready for more network frames */
					netif_wake_queue(priv->ndev);

					/* reset the full flag */
					priv->data_low_tx_full = 0;
				}
			}
		}

		if (reg & ISL38XX_INT_IDENT_INIT) {
			/* Device has been initialized */
#if VERBOSE > SHOW_ERROR_MESSAGES
			DEBUG(SHOW_TRACING,
			      "IRQ: Init flag, device initialized \n");
#endif

			/* initialize MIB ojects */
			schedule_work (&priv->mib_init_work);
		}

		if (reg & ISL38XX_INT_IDENT_SLEEP) {
			/* Device intends to move to powersave state */
#if VERBOSE > SHOW_ERROR_MESSAGES
			DEBUG(SHOW_TRACING, "IRQ: Sleep flag \n");
#endif
			isl38xx_handle_sleep_request(priv->control_block,
						     &priv->powerstate,
						     priv->device_base);
		}

		if (reg & ISL38XX_INT_IDENT_WAKEUP) {
			/* Device has been woken up to active state */
#if VERBOSE > SHOW_ERROR_MESSAGES
			DEBUG(SHOW_TRACING, "IRQ: Wakeup flag \n");
#endif

			isl38xx_handle_wakeup(priv->control_block,
					      &priv->powerstate,
					      priv->device_base);
		}
	}
	/* unlock the interrupt handler */
	spin_unlock(&priv->slock);

	return IRQ_HANDLED;
}

/******************************************************************************
    Network Interface Control & Statistical functions
******************************************************************************/
int
islpci_open(struct net_device *ndev)
{
	printk(KERN_DEBUG "%s: islpci_open()\n", ndev->name);

#ifdef WDS_LINKS
	open_wds_links(&ndev->priv->wdsp);
#endif
	netif_start_queue(ndev);
/*      netif_mark_up( ndev ); */

	return 0;
}

int
islpci_close(struct net_device *ndev)
{
	printk(KERN_DEBUG "%s: islpci_close ()\n", ndev->name);

#ifdef WDS_LINKS
	close_wds_links(&ndev->priv->wdsp);
#endif
	netif_stop_queue(ndev);
/*      netif_mark_down( ndev ); */

	return 0;
}

int
islpci_upload_fw(islpci_private *priv)
{
	printk(KERN_DEBUG "%s: uploading firmware...\n", priv->ndev->name);

	if (isl38xx_upload_firmware(priv->firmware,
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,5,75))
				    &priv->pdev->dev,
#else
				    pci_name(priv->pdev),
#endif
				    priv->device_base,
				    priv->device_host_address) == -1) {
		/* error uploading the firmware */
		printk(KERN_ERR "%s: could not upload firmware ('%s')\n",
		       priv->ndev->name, priv->firmware);

		return -1;
	}

	printk(KERN_DEBUG "%s: firmware upload & reset successful!\n",
	       priv->ndev->name);
	return 0;
}

int
islpci_reset(islpci_private *priv)
{
	islpci_qentry_t *entry;
	unsigned counter;

	printk(KERN_DEBUG "%s: resetting device...\n", priv->ndev->name);

	/* disable all device interrupts in case they weren't */
	isl38xx_disable_interrupts(priv->device_base);

	/* flush all management queues */
	while (!islpci_queue_get(&priv->mgmt_tx_shadowq, &entry))
		islpci_queue_put(&priv->mgmt_tx_freeq, entry);
	priv->index_mgmt_tx = 0;

	while (!islpci_queue_get(&priv->mgmt_rx_shadowq, &entry))
		islpci_queue_put(&priv->mgmt_rx_freeq, entry);
	priv->index_mgmt_rx = 0;

	while (!islpci_queue_get(&priv->ioctl_rx_queue, &entry))
		islpci_queue_put(&priv->mgmt_rx_freeq, entry);
	while (!islpci_queue_get(&priv->trap_rx_queue, &entry))
		islpci_queue_put(&priv->mgmt_rx_freeq, entry);
	while (!islpci_queue_get(&priv->pimfor_rx_queue, &entry))
		islpci_queue_put(&priv->mgmt_rx_freeq, entry);

	/* while at it, let's verify we don't have some leak... */
	if ((counter = islpci_queue_size(&priv->mgmt_rx_freeq))
	    != MGMT_RX_FRAME_COUNT)
		printk(KERN_CRIT
		       "%s: reclaimed rx frame count not correct (%d != %d)!\n",
		       priv->ndev->name, counter, MGMT_RX_FRAME_COUNT);

	if ((counter = islpci_queue_size(&priv->mgmt_tx_freeq))
	    != MGMT_TX_FRAME_COUNT)
		printk(KERN_CRIT
		       "%s: reclaimed tx frame count not correct (%d != %d)!\n",
		       priv->ndev->name, counter, MGMT_TX_FRAME_COUNT);

	/* clear the indexes in the frame pointer */
	for (counter = 0; counter < ISL38XX_CB_QCOUNT; counter++) {
		priv->control_block->driver_curr_frag[counter] = cpu_to_le32(0);
		priv->control_block->device_curr_frag[counter] = cpu_to_le32(0);
	}

	/* fill the shadow rx queues of both the management and data queues */
	for (counter = 0; counter < ISL38XX_CB_MGMT_QSIZE; counter++) {
		/* get an entry from the freeq and put it in the shadow queue */
		if (!islpci_queue_get(&priv->mgmt_rx_freeq, &entry))
			islpci_queue_put(&priv->mgmt_rx_shadowq, entry);
		else
			BUG();

		priv->control_block->rx_data_mgmt[counter].address =
		    cpu_to_le32(entry->dev_address);
	}			/* for */

	for (counter = 0; counter < ISL38XX_CB_RX_QSIZE; counter++) {
		priv->control_block->rx_data_low[counter].address =
		    cpu_to_le32((u32) priv->pci_map_rx_address[counter]);
	}

	BUG_ON(islpci_queue_size(&priv->mgmt_rx_shadowq) !=
	       ISL38XX_CB_MGMT_QSIZE);

	/* since the receive queues are filled with empty fragments, now we can
	 * set the corresponding indexes in the Control Block */
	priv->control_block->driver_curr_frag[ISL38XX_CB_RX_DATA_LQ] =
	    cpu_to_le32(ISL38XX_CB_RX_QSIZE);
	priv->control_block->driver_curr_frag[ISL38XX_CB_RX_MGMTQ] =
	    cpu_to_le32(ISL38XX_CB_MGMT_QSIZE);

	/* reset the remaining real index registers and full flags */
	priv->free_data_rx = 0;
	priv->free_data_tx = 0;
	priv->data_low_tx_full = 0;

	priv->powerstate = ISL38XX_PSM_ACTIVE_STATE;

	/* now that the data structures are cleaned up, upload
	   firmware and reset interface */
	return islpci_upload_fw(priv);
}

struct net_device_stats *
islpci_statistics(struct net_device *ndev)
{
	islpci_private *priv = ndev->priv;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_FUNCTION_CALLS, "islpci_statistics \n");
#endif

	return &priv->statistics;
}

/******************************************************************************
    Network device configuration functions
******************************************************************************/
int
islpci_alloc_memory(islpci_private *priv)
{
	int counter;
	long allocated, assigned;
	void *queue_base_addr;
	dma_addr_t queue_bus_addr;
	islpci_qentry_t *pointerq;

#if VERBOSE > SHOW_ERROR_MESSAGES
	printk(KERN_DEBUG "islpci_alloc_memory\n");
#endif

	/* remap the PCI device base address to accessable */
	if (!(priv->device_base =
	      ioremap(pci_resource_start(priv->pdev, 0),
		      ISL38XX_PCI_MEM_SIZE))) {
		/* error in remapping the PCI device memory address range */
		printk(KERN_ERR "PCI memory remapping failed \n");
		return -1;
	}

	/* memory layout for consistent DMA region:
	 *
	 * Area 1: Control Block for the device interface
	 * Area 2: Power Save Mode Buffer for temporary frame storage. Be aware that
	 *         the number of supported stations in the AP determines the minimal
	 *         size of the buffer !
	 * Area 3: Driver Queuing System resources 
	 */

	/* set the allocated parameter for single alloc */
	allocated = HOST_MEM_BLOCK;

	/* perform the allocation */
	priv->driver_mem_address = pci_alloc_consistent(priv->pdev, allocated,
							&priv->
							device_host_address);

	if (!priv->driver_mem_address) {
		/* error allocating the block of PCI memory */
		printk(KERN_ERR "%s: could not allocate DMA memory, aborting!",
		       "prism54");
		return -1;
	}

	/* assign the Control Block to the first address of the allocated area */
	assigned = 0;
	priv->control_block =
	    (isl38xx_control_block *) priv->driver_mem_address;
	assigned += CONTROL_BLOCK_SIZE;

	/* set the Power Save Buffer pointer directly behind the CB */
	priv->device_psm_buffer = priv->device_host_address + assigned;
	assigned += PSM_BUFFER_SIZE;

	/* set the working pointers */
	queue_base_addr = priv->driver_mem_address + assigned;
	queue_bus_addr = priv->device_host_address + assigned;

	/* assign memory for the management frames  */
	for (counter = 0; counter < MGMT_FRAME_COUNT; counter++) {
		/* check whether the frame can be stored in the current page */
		BUG_ON(allocated - assigned < MGMT_FRAME_SIZE);

		/* configure the queue entry at the beginning of the memory block */
		pointerq = &priv->_queue_slots[counter];
		pointerq->host_address = (u32) queue_base_addr;
		pointerq->dev_address = (u32) queue_bus_addr;
		pointerq->size = MGMT_FRAME_SIZE;
		pointerq->fragments = 1;

		/* add the configured entry to the management tx/rx root queues */
		if (counter < MGMT_TX_FRAME_COUNT)
			islpci_queue_put(&priv->mgmt_tx_freeq, pointerq);
		else
			islpci_queue_put(&priv->mgmt_rx_freeq, pointerq);

		/* update the working pointers and registers */
		queue_base_addr += MGMT_FRAME_SIZE;
		queue_bus_addr += MGMT_FRAME_SIZE;
		assigned += MGMT_FRAME_SIZE;
	}

	/* now get the data rx skb's */
	memset(priv->data_low_rx, 0, sizeof (priv->data_low_rx));
	memset(priv->pci_map_rx_address, 0, sizeof (priv->pci_map_rx_address));

	for (counter = 0; counter < ISL38XX_CB_RX_QSIZE; counter++) {
		struct sk_buff *skb;

		/* allocate an sk_buff for received data frames storage
		 * each frame on receive size consists of 1 fragment
		 * include any required allignment operations */
		if (!(skb = dev_alloc_skb(MAX_FRAGMENT_SIZE_RX + 2))) {
			/* error allocating an sk_buff structure elements */
			printk(KERN_ERR "Error allocating skb.\n");
			break;
		}
		/* add the new allocated sk_buff to the buffer array */
		priv->data_low_rx[counter] = skb;

		/* map the allocated skb data area to pci */
		priv->pci_map_rx_address[counter] =
		    pci_map_single(priv->pdev, (void *) skb->data,
				   MAX_FRAGMENT_SIZE_RX + 2, PCI_DMA_FROMDEVICE);
		if (!priv->pci_map_rx_address[counter]) {
			/* error mapping the buffer to device
			   accessable memory address */
			printk(KERN_ERR "failed to map skb DMA'able\n");
			break;
		}
	}

	if (counter != ISL38XX_CB_RX_QSIZE) {
		/* loop was left prematurely, do cleanup */
		islpci_free_memory(priv);
		return -1;
	}

	return 0;
}

int
islpci_free_memory(islpci_private *priv)
{
	int counter;

	if (priv->device_base)
		iounmap(priv->device_base);
	priv->device_base = 0;

	/* free consistent DMA area... */
	if (priv->driver_mem_address)
		pci_free_consistent(priv->pdev, HOST_MEM_BLOCK,
				    priv->driver_mem_address,
				    priv->device_host_address);

	/* clear some dangling pointers */
	memset(priv->_queue_slots, 0, sizeof (priv->_queue_slots));
	priv->driver_mem_address = 0;
	priv->device_host_address = 0;
	priv->device_psm_buffer = 0;
	priv->control_block = 0;

	/* clean up data rx buffers */
	for (counter = 0; counter < ISL38XX_CB_RX_QSIZE; counter++) {
		if (priv->pci_map_rx_address[counter])
			pci_unmap_single(priv->pdev,
					 priv->pci_map_rx_address[counter],
					 MAX_FRAGMENT_SIZE_RX + 2,
					 PCI_DMA_FROMDEVICE);
		priv->pci_map_rx_address[counter] = 0;

		if (priv->data_low_rx[counter])
			dev_kfree_skb(priv->data_low_rx[counter]);
		priv->data_low_rx[counter] = 0;
	}

	return 0;
}

#if 0
static void
islpci_set_multicast_list(struct net_device *dev)
{
	/* put device into promisc mode and let network layer handle it */
}
#endif

/* requires islpci_alloc_memory() to have been called already! */
struct net_device *
islpci_setup(struct pci_dev *pdev)
{
	islpci_private *priv;
	struct net_device *ndev = alloc_etherdev(sizeof (islpci_private));

	if (!ndev)
		return ndev;

	SET_MODULE_OWNER(ndev);
	pci_set_drvdata(pdev, ndev);
#if defined(SET_NETDEV_DEV)
	SET_NETDEV_DEV(ndev, &pdev->dev);
#endif

	/* setup the structure members */
	ndev->base_addr = pci_resource_start(pdev, 0);
	ndev->irq = pdev->irq;

	/* initialize the function pointers */
	ndev->open = &islpci_open;
	ndev->stop = &islpci_close;
	ndev->get_stats = &islpci_statistics;
	ndev->get_wireless_stats = &prism54_get_wireless_stats;
	ndev->do_ioctl = &prism54_ioctl;
#if WIRELESS_EXT > 12
	ndev->wireless_handlers =
	    (struct iw_handler_def *) &prism54_handler_def;
#endif				/* WIRELESS_EXT > 12 */

	ndev->hard_start_xmit = &islpci_eth_transmit;
	/* ndev->set_multicast_list = &islpci_set_multicast_list; */
	ndev->type = ARPHRD_ETHER;
	ndev->addr_len = ETH_ALEN;

#ifdef HAVE_TX_TIMEOUT
	ndev->watchdog_timeo = ISLPCI_TX_TIMEOUT;
	ndev->tx_timeout = &islpci_eth_tx_timeout;
#endif

	/* allocate a private device structure to the network device  */
	priv = ndev->priv;
	priv->ndev = ndev;
	priv->pdev = pdev;

	/* init the MIB backup */
	prism54_mib_init_defaults (&priv->mib);

	/* save the start and end address of the PCI memory area */
	ndev->mem_start = (unsigned long) priv->device_base;
	ndev->mem_end = ndev->mem_start + ISL38XX_PCI_MEM_SIZE;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "PCI Memory remapped to 0x%p\n", priv->device_base);
#endif

	/* initialize all the queues in the private configuration structure */
	islpci_queue_init(&priv->mgmt_tx_freeq);
	islpci_queue_init(&priv->mgmt_rx_freeq);
	islpci_queue_init(&priv->mgmt_tx_shadowq);
	islpci_queue_init(&priv->mgmt_rx_shadowq);
	islpci_queue_init(&priv->ioctl_rx_queue);
	islpci_queue_init(&priv->trap_rx_queue);
	islpci_queue_init(&priv->pimfor_rx_queue);

	/* init the queue read locks, process wait counter */
	init_waitqueue_head(&priv->ioctl_reader_wqueue);
	atomic_set(&priv->ioctl_reader_count, 0);
	sema_init(&priv->ioctl_reader_sem, 1);
	spin_lock_init(&priv->slock);
	priv->flags = 0;

	/* initialize workqueue's */
	INIT_WORK(&priv->mib_init_work, 
		  (void (*)(void *)) prism54_mib_init, priv);
	
	/* allocate various memory areas */
	if (islpci_alloc_memory(priv))
		goto do_free_netdev;

	/* set the power save state */
	priv->powerstate = ISL38XX_PSM_ACTIVE_STATE;

	/* select the firmware file depending on the device id */
	switch (pdev->device) {
	case PCIDEVICE_ISL3890:
	case PCIDEVICE_3COM6001:
		strcpy(priv->firmware, ISL3890_IMAGE_FILE);
		break;
	case PCIDEVICE_ISL3877:
		strcpy(priv->firmware, ISL3877_IMAGE_FILE);
		break;

	default:
		BUG();		/* someone forgot to add a new case entry... */
		break;
	}

	if (register_netdev(ndev)) {
		DEBUG(SHOW_ERROR_MESSAGES,
		      "ERROR: register_netdev() failed \n");
		goto do_islpci_free_memory;
	}

	return ndev;

      do_islpci_free_memory:
	islpci_free_memory(priv);
      do_free_netdev:
	pci_set_drvdata(pdev, 0);
	free_netdev(ndev);
	priv = 0;
	return NULL;
}
