/*  $Header: /var/lib/cvs/prism54-ng/ksrc/islpci_dev.h,v 1.23 2003/11/12 12:20:15 ajfa Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc. 
 *  Copyright (C) 2003 Herbert Valerio Riedel <hvr@gnu.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _ISLPCI_DEV_H
#define _ISLPCI_DEV_H

#include <linux/version.h>
#include <linux/netdevice.h>
#include <linux/wireless.h>
#include <linux/list.h>

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,41)
# include <linux/workqueue.h>
#else
# include <linux/tqueue.h>
# define work_struct tq_struct
# define INIT_WORK INIT_TQUEUE
# define schedule_work schedule_task
#endif

#include "isl_38xx.h"
#ifdef WDS_LINKS
#include "isl_wds.h"
#endif

#include "bloboid.h"

typedef struct {
	struct list_head _list;
	u32 host_address;	/* mem address for the host */
	u32 dev_address;	/* mem address for the device */
	u16 size;		/* size of memory block or queue */
	u16 fragments;		/* nr of fragments in data part */
} islpci_qentry_t;

typedef struct {
	struct list_head list;
	int size;		/* size of queue */
	spinlock_t _slock;	/* mutex */
} islpci_queue_t;

/* MIB mirror backup */
struct islpci_mib {
	u32 mode;		/* OID_INL_MODE */
	struct obj_ssid ssid;	/* DOT11_OID_SSID */
	struct obj_key defkey[4];
	u32 frequency;
	u32 channel;
	u32 mlmeautolevel;	/* DOT11_OID_MLMEAUTOLEVEL */
	u32 authenable;
	u32 privacyinvoked;
	u32 bsstype;
	u32 exunencrypted;
	u32 conformance;
	u32 dot1xenable;
};

typedef struct _islpci_private {
	struct islpci_mib mib;

	/* prism54_mib_init */
	struct work_struct mib_init_work;

	/* some info about card state that we want to read and set atomically.
	 * We will use a second flag if we want iwspy support. Maybe it can feat in
	 * your state field hvr ?
	 */
	unsigned long flags;
#define FLAG_PROMISC	1	/* monitor mode set */

	/* PCI bus allocation & configuration members */
	struct pci_dev *pdev;	/* PCI structure information */
	u32 pci_state[16];	/* used for suspend/resume */
	char firmware[33];

	void *device_base;	/* ioremapped device base address */

	/* consistent DMA region */
	void *driver_mem_address;	/* base DMA address */
	dma_addr_t device_host_address;	/* base DMA address (bus address) */
	dma_addr_t device_psm_buffer;	/* host memory for PSM buffering (bus address) */

	/* our network_device structure  */
	struct net_device *ndev;

	/* device queue interface members */
	volatile struct isl38xx_cb *control_block;	/* device control block 
							   (== driver_mem_address!) */

	u32 index_mgmt_rx;	/* real index mgmt rx queue */
	u32 index_mgmt_tx;	/* read index mgmt tx queue */
	u32 free_data_rx;	/* free pointer data rx queue */
	u32 free_data_tx;	/* free pointer data tx queue */
	u32 data_low_tx_full;	/* full detected flag */

	/* root free queue definitions */
	islpci_qentry_t _queue_slots[MGMT_FRAME_COUNT];	/* all available queue slots */

	islpci_queue_t mgmt_tx_freeq;	/* queue entry container mgmt tx */
	islpci_queue_t mgmt_rx_freeq;	/* queue entry container mgmt rx */

	/* driver queue definitions */
	islpci_queue_t mgmt_tx_shadowq;	/* shadow queue mgmt tx */
	islpci_queue_t mgmt_rx_shadowq;	/* shadow queue mgmt rx */
	islpci_queue_t ioctl_rx_queue;	/* ioctl rx queue */
	islpci_queue_t trap_rx_queue;	/* TRAP pimfor rx queue */
	islpci_queue_t pimfor_rx_queue;	/* PIMFOR generated rx queue */

	struct sk_buff *data_low_tx[ISL38XX_CB_TX_QSIZE];
	struct sk_buff *data_low_rx[ISL38XX_CB_RX_QSIZE];
	dma_addr_t pci_map_tx_address[ISL38XX_CB_TX_QSIZE];
	dma_addr_t pci_map_rx_address[ISL38XX_CB_RX_QSIZE];

	/* driver network interface members */
	struct net_device_stats statistics;
	struct iw_statistics iwstatistics;

	/* used by islpci_rec islpci_mgt_receive/pci_mgt_response */
	wait_queue_head_t ioctl_reader_wqueue;
	struct semaphore ioctl_reader_sem;
	atomic_t ioctl_reader_count;

	/* driver operational members */
	int powerstate;		/* device power save state */

	/* state machine */
	int hw_unavailable;	/* hw unavailable for low-level
				   access? (i.e. hardware was taken
				   away or is in suspend state */

	/* some states might not be superflous and may be removed when
	   design is finalized (hvr) */
	enum {
		PRV_STATE_OFF = 0,	/* this means hw_unavailable is != 0 */
		PRV_STATE_PREBOOT,	/* we are in a pre-boot state (empty RAM) */
		PRV_STATE_BOOT,	/* boot state (fw upload, run fw, reset if) */
		PRV_STATE_PREINIT,	/* pre-init state */
		PRV_STATE_INIT,	/* init state (restore MIB backup to device) */
		PRV_STATE_READY,	/* driver&device are in operational state */
		PRV_STATE_SLEEP	/* device in sleep mode */
	} state;

	spinlock_t slock;	/* spinlock */

#ifdef WDS_LINKS
	struct wds_priv wdsp;
#endif
} islpci_private;

#define ISLPCI_TX_TIMEOUT               (2*HZ)

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,5,75))
irqreturn_t islpci_interrupt(int, void *, struct pt_regs *);
#else
void islpci_interrupt(int, void *, struct pt_regs *);
#endif

int islpci_open(struct net_device *);
int islpci_close(struct net_device *);
int islpci_reset(islpci_private *);
int islpci_upload_fw(islpci_private *);
struct net_device_stats *islpci_statistics(struct net_device *);

int islpci_alloc_memory(islpci_private *);
int islpci_free_memory(islpci_private *);
struct net_device *islpci_setup(struct pci_dev *);

#endif				/* _ISLPCI_DEV_H */
