/*  $Header: /var/lib/cvs/prism54-ng/ksrc/islpci_eth.c,v 1.21 2004/01/08 04:14:40 mcgrof Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/version.h>
#include <linux/module.h>

#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>

#include "isl_38xx.h"
#include "islpci_eth.h"
#include "islpci_mgt.h"

#ifdef CONFIG_PRISM54_WDS
#include "isl_wds.h"
#endif

/******************************************************************************
    Network Interface functions
******************************************************************************/
void
islpci_eth_cleanup_transmit(islpci_private *priv,
			    volatile isl38xx_control_block *control_block)
{
	struct sk_buff *skb;
	u32 index;

	/* compare the control block read pointer with the free pointer */
	while (priv->free_data_tx !=
	       le32_to_cpu(control_block->
			   device_curr_frag[ISL38XX_CB_TX_DATA_LQ])) {
		/* read the index of the first fragment to be freed */
		index = priv->free_data_tx % ISL38XX_CB_TX_QSIZE;

		/* check for holes in the arrays caused by multi fragment frames 
		 * searching for the last fragment of a frame */
		if (priv->pci_map_tx_address[index] != (dma_addr_t) NULL) {
			/* entry is the last fragment of a frame
			 * free the skb structure and unmap pci memory */
			skb = priv->data_low_tx[index];

#if VERBOSE > SHOW_ERROR_MESSAGES
			DEBUG(SHOW_TRACING,
			      "cleanup skb %p skb->data %p skb->len %u truesize %u\n ",
			      skb, skb->data, skb->len, skb->truesize);
#endif

			pci_unmap_single(priv->pdev,
					 priv->pci_map_tx_address[index],
					 skb->len, PCI_DMA_TODEVICE);
			dev_kfree_skb_irq(skb);
		}
		/* increment the free data low queue pointer */
		priv->free_data_tx++;
	}
}

int
islpci_eth_transmit(struct sk_buff *skb, struct net_device *ndev)
{
	islpci_private *priv = ndev->priv;
	volatile isl38xx_control_block *control_block = priv->control_block;
	u32 index, counter, pci_map_address;
	int fragments;
	int frame_size;
	int offset;
	struct sk_buff *newskb;
	int newskb_offset;
	unsigned long flags;
	unsigned char wds_mac[6];
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_FUNCTION_CALLS, "islpci_eth_transmit \n");
#endif

	/* lock the driver code */
	spin_lock_irqsave(&priv->slock, flags);

	/* check whether the deviec is running
	 *      if( !netif_running( ndev ) )
	 *      {
	 *              DEBUG( SHOW_ERROR_MESSAGES, "%s: Tx on stopped device!\n",
	 *                     ndev->name);
	 *              return 1;
	 *      } */

/*      tx_timeout_check( ndev, islpci_eth_tx_timeout );
 *      skb_tx_check( ndev, skb ); */

	/* determine the amount of fragments needed to store the frame */
#ifdef CONFIG_PRISM54_WDS
	memset(wds_mac, 0, 6);
	check_skb_for_wds(skb, wds_mac);
#ifdef ISLPCI_ETH_DEBUG
	if (wds_mac[0] || wds_mac[1] || wds_mac[2] || wds_mac[3] || wds_mac[4]
	    || wds_mac[5]) {
		printk
		    ("islpci_eth_transmit:check_skb_for_wds ! %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
		     wds_mac[0], wds_mac[1], wds_mac[2], wds_mac[3], wds_mac[4],
		     wds_mac[5]);
	} else {
		printk
		    ("islpci_eth_transmit:check_skb_for_wds %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
		     wds_mac[0], wds_mac[1], wds_mac[2], wds_mac[3], wds_mac[4],
		     wds_mac[5]);
	}
#endif
#endif

	frame_size = skb->len < ETH_ZLEN ? ETH_ZLEN : skb->len;
	if (init_wds)
		frame_size += 6;
	fragments = frame_size / MAX_FRAGMENT_SIZE;
	fragments += frame_size % MAX_FRAGMENT_SIZE == 0 ? 0 : 1;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "Fragments needed for frame %i\n", fragments);
#endif

	/* check whether the destination queue has enough fragments for the frame */
	if (ISL38XX_CB_TX_QSIZE - isl38xx_in_queue(control_block,
						   ISL38XX_CB_TX_DATA_LQ) <
	    fragments) {
		/* Error, cannot add the frame because the queue is full */
		DEBUG(SHOW_ERROR_MESSAGES,
		      "Error: Queue [%i] not enough room\n",
		      ISL38XX_CB_TX_DATA_LQ);

		/* trigger the device by setting the Update bit in the Device Int reg */
		isl38xx_w32_flush(priv->device_base, ISL38XX_DEV_INT_UPDATE,
				  ISL38XX_DEV_INT_REG);
		udelay(ISL38XX_WRITEIO_DELAY);

		/* unlock the driver code */
		spin_unlock_irqrestore(&priv->slock, flags);
		return -EBUSY;
	}
	/* Check alignment and WDS frame formatting. The start of the packet should
	 * be aligned on a 4-byte boundary. If WDS is enabled add another 6 bytes
	 * and add WDS address information */
	if (((int) skb->data & 0x03) | init_wds) {
		/* get the number of bytes to add and re-allign */
		offset = (4 - (int) skb->data) & 0x03;
		offset += init_wds ? 6 : 0;

		/* check whether the current skb can be used  */
		if (!skb_cloned(skb) && (skb_tailroom(skb) >= offset)) {
			unsigned char *src = skb->data;

#if VERBOSE > SHOW_ERROR_MESSAGES
			DEBUG(SHOW_TRACING, "skb offset %i wds %i\n", offset,
			      init_wds);
#endif

			/* allign the buffer on 4-byte boundary */
			skb_reserve(skb, (4 - (int) skb->data) & 0x03);
			if (init_wds) {
				/* wds requires an additional address field of 6 bytes */
				skb_put(skb, 6);
#ifdef ISLPCI_ETH_DEBUG
				printk("islpci_eth_transmit:wds_mac\n");
#endif
				memmove(skb->data + 6, src, skb->len);
				memcpy(skb->data, wds_mac, 6);
			} else {
				memmove(skb->data, src, skb->len);
			}

#if VERBOSE > SHOW_ERROR_MESSAGES
			DEBUG(SHOW_TRACING, "memmove %p %p %i \n", skb->data,
			      src, skb->len);
#endif
		} else {
			newskb =
			    dev_alloc_skb(init_wds ? skb->len + 6 : skb->len);
			newskb_offset = (4 - (int) newskb->data) & 0x03;

			/* Check if newskb->data is aligned */
			if (newskb_offset)
				skb_reserve(newskb, newskb_offset);

			skb_put(newskb, init_wds ? skb->len + 6 : skb->len);
			if (init_wds) {
				memcpy(newskb->data + 6, skb->data, skb->len);
				memcpy(newskb->data, wds_mac, 6);
#ifdef ISLPCI_ETH_DEBUG
				printk("islpci_eth_transmit:wds_mac\n");
#endif
			} else
				memcpy(newskb->data, skb->data, skb->len);

#if VERBOSE > SHOW_ERROR_MESSAGES
			DEBUG(SHOW_TRACING, "memcpy %p %p %i wds %i\n",
			      newskb->data, skb->data, skb->len, init_wds);
#endif

			newskb->dev = skb->dev;
			dev_kfree_skb(skb);
			skb = newskb;
		}
	}
	/* display the buffer contents for debugging */
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_BUFFER_CONTENTS, "\ntx %p ", skb->data);
	display_buffer((char *) skb->data, skb->len);
#endif

	/* map the skb buffer to pci memory for DMA operation */
	pci_map_address = pci_map_single(priv->pdev,
					 (void *) skb->data, skb->len,
					 PCI_DMA_TODEVICE);
	if (pci_map_address == (dma_addr_t) NULL) {
		/* error mapping the buffer to device accessable memory address */
		DEBUG(SHOW_ERROR_MESSAGES, "Error map DMA addr, data lost\n");

		/* free the skbuf structure before aborting */
		dev_kfree_skb(skb);

		/* unlock the driver code */
		spin_unlock_irqrestore(&priv->slock, flags);
		return -EIO;
	}
	/* place each fragment in the control block structure and store in the last
	 * needed fragment entry of the pci_map_tx_address and data_low_tx arrays 
	 * the skb frame information */
	for (counter = 0; counter < fragments; counter++) {
		volatile isl38xx_fragment *fragment;

		/* get a pointer to the target control block fragment */
		index =
		    (counter +
		     le32_to_cpu(control_block->
				 driver_curr_frag[ISL38XX_CB_TX_DATA_LQ])) %
		    ISL38XX_CB_TX_QSIZE;

		fragment = &control_block->tx_data_low[index];

		/* check whether this frame fragment is the last one */
		if (counter == fragments - 1) {
			/* the fragment is the last one, add the streaming DMA mapping for 
			 * proper PCI bus operation */
			priv->pci_map_tx_address[index] =
			    (dma_addr_t) pci_map_address;

			/* store the skb address for future freeing  */
			priv->data_low_tx[index] = skb;
		} else {
			/* the fragment will be followed by more fragments
			 * clear the pci_map_tx_address and data_low_tx entries */
			priv->pci_map_tx_address[index] = (dma_addr_t) NULL;
			priv->data_low_tx[index] = NULL;
		}

		/* set the proper fragment start address and size information */
		fragment->address = cpu_to_le32(pci_map_address +
						counter * MAX_FRAGMENT_SIZE);
		fragment->size = cpu_to_le16(frame_size > MAX_FRAGMENT_SIZE ?
					     MAX_FRAGMENT_SIZE : frame_size);
		fragment->flags =
		    cpu_to_le16(frame_size > MAX_FRAGMENT_SIZE ? 1 : 0);
		frame_size -= MAX_FRAGMENT_SIZE;
	}

	/* ready loading all fragements to the control block and setup pci mapping
	 * update the control block interface information */
	wmb();
	add_le32p((u32 *) & control_block->
		  driver_curr_frag[ISL38XX_CB_TX_DATA_LQ], fragments);

	/* check whether the data tx queue is full now */
	if (ISL38XX_CB_TX_QSIZE - isl38xx_in_queue(control_block,
						   ISL38XX_CB_TX_DATA_LQ) <
	    ISL38XX_MIN_QTHRESHOLD) {
		/* stop the transmission of network frames to the driver */
		netif_stop_queue(ndev);

		/* set the full flag for the data low transmission queue */
		priv->data_low_tx_full = 1;
	}

	/* trigger the device */
	islpci_trigger(priv);

	/* set the transmission time */
	ndev->trans_start = jiffies;
	priv->statistics.tx_packets++;
	priv->statistics.tx_bytes += skb->len;

	/* unlock the driver code */
	spin_unlock_irqrestore(&priv->slock, flags);
	return 0;
}

int
islpci_eth_receive(islpci_private *priv)
{
	struct net_device *ndev = priv->ndev;
	volatile isl38xx_control_block *control_block = priv->control_block;
#ifdef CONFIG_PRISM54_WDS
	struct net_device *wds_dev = NULL;
	struct wds_net_local *wds_lp;
	struct wds_priv *wdsp = &priv->wdsp;
#endif
	struct sk_buff *skb;
	u16 size;
	u32 index, offset;
	unsigned char *src;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_FUNCTION_CALLS, "islpci_eth_receive \n");
#endif

	/* the device has written an Ethernet frame in the data area
	 * of the sk_buff without updating the structure, do it now */
	index = priv->free_data_rx % ISL38XX_CB_RX_QSIZE;
	size = le16_to_cpu(control_block->rx_data_low[index].size);
	skb = priv->data_low_rx[index];
	offset = ((u32) le32_to_cpu(control_block->rx_data_low[index].address) -
		  (u32) skb->data) & 3;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING,
	      "frq->addr %x skb->data %p skb->len %u offset %u truesize %u\n ",
	      control_block->rx_data_low[priv->free_data_rx].address, skb->data,
	      skb->len, offset, skb->truesize);
#endif

	/* delete the streaming DMA mapping before processing the skb */
	pci_unmap_single(priv->pdev,
			 priv->pci_map_rx_address[index],
			 MAX_FRAGMENT_SIZE_RX + 2, PCI_DMA_FROMDEVICE);

	/* update the skb structure and allign the buffer */
	skb_put(skb, size);
	if (offset) {
		/* shift the buffer allocation offset bytes to get the right frame */
		skb_pull(skb, 2);
		skb_put(skb, 2);
	}
#if VERBOSE > SHOW_ERROR_MESSAGES
	/* display the buffer contents for debugging */
	DEBUG(SHOW_BUFFER_CONTENTS, "\nrx %p ", skb->data);
	display_buffer((char *) skb->data, skb->len);
#endif

	/* check whether WDS is enabled and whether the data frame is a WDS frame */

	if (init_wds) {
		/* WDS enabled, check for the wds address on the first 6 bytes of the buffer */
#ifdef CONFIG_PRISM54_WDS
		wds_dev = wds_find_device(skb->data, wdsp);
#ifdef ISLPCI_ETH_DEBUG
		if (wds_dev) {
			printk
			    ("islpci_eth_receive:wds_find_device ! %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
			     skb->data[0], skb->data[1], skb->data[2],
			     skb->data[3], skb->data[4], skb->data[5]);
		} else {
			printk
			    ("islpci_eth_receive:wds_find_device %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
			     skb->data[0], skb->data[1], skb->data[2],
			     skb->data[3], skb->data[4], skb->data[5]);
		}
#endif
#endif
		src = skb->data + 6;
		memmove(skb->data, src, skb->len - 6);
		skb_trim(skb, skb->len - 6);
	}
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "Fragment size %i in skb at %p\n", size, skb);
	DEBUG(SHOW_TRACING, "Skb data at %p, length %i\n", skb->data, skb->len);

	/* display the buffer contents for debugging */
	DEBUG(SHOW_BUFFER_CONTENTS, "\nrx %p ", skb->data);
	display_buffer((char *) skb->data, skb->len);
#endif

	/* do some additional sk_buff and network layer parameters */
	skb->dev = ndev;
	skb->protocol = eth_type_trans(skb, ndev);
	skb->ip_summed = CHECKSUM_NONE;
	priv->statistics.rx_packets++;
	priv->statistics.rx_bytes += size;

	/* take care of monitor mode */
	if (priv->mib.mode == INL_MODE_PROMISCUOUS) {
		/* The card reports full 802.11 packets but with a 6 bytes header */
		skb_pull(skb, 6);
		skb->protocol = htons(ETH_P_802_2);
		skb->mac.raw = skb->data;
		skb->pkt_type = PACKET_OTHERHOST;
	}
#ifdef CONFIG_PRISM54_WDS
	if (wds_dev != NULL) {
		wds_lp = (struct wds_net_local *) wds_dev->priv;
		wds_lp->stats.rx_packets++;
		wds_lp->stats.rx_bytes += skb->len;
		skb->dev = wds_dev;
	}
#endif
	/* deliver the skb to the network layer */
#ifdef ISLPCI_ETH_DEBUG
	printk
	    ("islpci_eth_receive:netif_rx %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
	     skb->data[0], skb->data[1], skb->data[2], skb->data[3],
	     skb->data[4], skb->data[5]);
#endif
	netif_rx(skb);

	/* increment the read index for the rx data low queue */
	priv->free_data_rx++;

	/* add one or more sk_buff structures */
	while (index =
	       le32_to_cpu(control_block->
			   driver_curr_frag[ISL38XX_CB_RX_DATA_LQ]),
	       index - priv->free_data_rx < ISL38XX_CB_RX_QSIZE) {
		/* allocate an sk_buff for received data frames storage
		 * include any required allignment operations */
		if (skb = dev_alloc_skb(MAX_FRAGMENT_SIZE_RX + 2), skb == NULL) {
			/* error allocating an sk_buff structure elements */
			DEBUG(SHOW_ERROR_MESSAGES, "Error allocating skb \n");
			break;
		}
		/* store the new skb structure pointer */
		index = index % ISL38XX_CB_RX_QSIZE;
		priv->data_low_rx[index] = skb;

#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_TRACING,
		      "new alloc skb %p skb->data %p skb->len %u index %u truesize %u\n ",
		      skb, skb->data, skb->len, index, skb->truesize);
#endif

		/* set the streaming DMA mapping for proper PCI bus operation */
		priv->pci_map_rx_address[index] =
		    pci_map_single(priv->pdev, (void *) skb->data,
				   MAX_FRAGMENT_SIZE_RX + 2,
				   PCI_DMA_FROMDEVICE);
		if (priv->pci_map_rx_address[index] == (dma_addr_t) NULL) {
			/* error mapping the buffer to device accessable memory address */
			DEBUG(SHOW_ERROR_MESSAGES,
			      "Error mapping DMA address\n");

			/* free the skbuf structure before aborting */
			dev_kfree_skb((struct sk_buff *) skb);
			break;
		}
		/* update the fragment address */
		control_block->rx_data_low[index].address = cpu_to_le32((u32)
									priv->
									pci_map_rx_address
									[index]);

		/* increment the driver read pointer */
		add_le32p((u32 *) & control_block->
			  driver_curr_frag[ISL38XX_CB_RX_DATA_LQ], 1);
	}

	/* trigger the device */
	islpci_trigger(priv);

	return 0;
}

void
islpci_eth_tx_timeout(struct net_device *ndev)
{
	islpci_private *priv = ndev->priv;
	struct net_device_stats *statistics = &priv->statistics;

	/* increment the transmit error counter */
	statistics->tx_errors++;

#if 0
	/* don't do this here! we are not allowed to sleep since we are in interrupt context */
	if (islpci_reset(priv))
		printk(KERN_ERR "%s: error on TX timeout card reset!\n",
		       ndev->name);
#endif

	/* netif_wake_queue(ndev); */
	return;
}
