/*  $Header: /var/lib/cvs/prism54-ng/ksrc/islpci_mgt.c,v 1.31 2004/01/08 04:14:40 mcgrof Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/config.h>
#include <linux/netdevice.h>
#include <linux/module.h>

#include <asm/io.h>
#include <linux/if_arp.h>

#include "isl_38xx.h"
#include "islpci_mgt.h"
#include "isl_oid.h"		/* additional types and defs for isl38xx fw */
#include "isl_ioctl.h"

#ifdef CONFIG_PRISM54_WDS
# include "isl_wds.h"
# include <linux/compiler.h>
# ifdef __deprecated
int __deprecated
mgt_response(unsigned int src_id, unsigned int src_seq,
	     unsigned int dev_type, unsigned int dev_seq,
	     unsigned int op, int oid, void *data, unsigned long len)
{
	BUG();
	return 0;
}
# endif
#endif

#if WIRELESS_EXT > 12
#include <net/iw_handler.h>
#endif

/******************************************************************************
        Global variable definition section
******************************************************************************/
int pc_debug = VERBOSE;
MODULE_PARM(pc_debug, "i");

/******************************************************************************
    Driver general functions
******************************************************************************/
void
display_buffer(char *buffer, int length)
{
	if ((pc_debug & SHOW_BUFFER_CONTENTS) == 0)
		return;

	while (length > 0) {
		printk("[%02x]", *buffer & 255);
		length--;
		buffer++;
	}

	printk("\n");
}

void
print_frame(struct sk_buff *skb)
{
	int offset;

	printk("pr_frame: size %d, skb->data %p, skb %p\n",
	       skb->len, skb->data, skb);
	for (offset = 0; offset < skb->len; offset++)
		printk("%02X ", skb->data[offset]);
	printk("\n");
}

void
string_to_macaddress(unsigned char *string, unsigned char *mac)
{
	unsigned char *wp = mac, *rp;
	unsigned int value;
	unsigned int count = 0;

	for (rp = string; *rp != '\0'; rp++, wp++, count++) {
		if (count == 6)
			break;

		if (*rp >= '0' && *rp <= '9')
			value = *rp - '0';
		else if (*rp >= 'a' && *rp <= 'f')
			value = *rp - 'a' + 10;
		else if (*rp >= 'A' && *rp <= 'F')
			value = *rp - 'A' + 10;
		else
			break;

		value <<= 4;
		rp++;
		if (*rp >= '0' && *rp <= '9')
			value |= *rp - '0';
		else if (*rp >= 'a' && *rp <= 'f')
			value |= *rp - 'a' + 10;
		else if (*rp >= 'A' && *rp <= 'F')
			value |= *rp - 'A' + 10;
		else
			break;

		*wp = (unsigned char) value;
	}
}

/******************************************************************************
    Frame (re)format functions
******************************************************************************/
void
pimfor_encode_header(int operation, u32 oid,
		     int device_id, int flags, u32 length,
		     pimfor_header_t *header)
{
	/* byte oriented members */
	header->version = PIMFOR_VERSION;
	header->operation = operation;
	header->device_id = device_id;
	header->flags = flags;

	/* word oriented members with byte order depending on the flags */
	if (flags & PIMFOR_FLAG_LITTLE_ENDIAN) {
		/* use little endian coding */
		header->oid = cpu_to_le32(oid);
		header->length = cpu_to_le32(length);
	} else {
		/* use big endian coding */
		header->oid = cpu_to_be32(oid);
		header->length = cpu_to_be32(length);
	}
}

void
pimfor_decode_header(const pimfor_header_t *header, int *operation,
		     u32 * oid, int *device_id, int *flags, u32 * length)
{
	/* byte oriented members */
	*operation = header->operation;
	*device_id = header->device_id;
	*flags = header->flags;

	/* word oriented members with byte order depending on the flags */
	if (*flags & PIMFOR_FLAG_LITTLE_ENDIAN) {
		/* use little endian coding */
		*oid = le32_to_cpu(header->oid);
		*length = le32_to_cpu(header->length);
	} else {
		/* use big endian coding */
		*oid = be32_to_cpu(header->oid);
		*length = be32_to_cpu(header->length);
	}
}

/******************************************************************************
    Memory Management functions
******************************************************************************/
void
islpci_queue_init(islpci_queue_t *queue)
{
	INIT_LIST_HEAD(&queue->list);
	spin_lock_init(&queue->_slock);
	queue->size = 0;
}

int
islpci_queue_put(islpci_queue_t *queue, islpci_qentry_t *entry)
{
	unsigned long flags;

	/* the size and next members are both accessed from kernel
	 * code and from interrupt handler, disable the update
	 * interrupt as short as possible */

	BUG_ON(entry == 0);

	/*isl38xx_disable_update_interrupt(device); */
	spin_lock_irqsave(&queue->_slock, flags);

	/* check whether the queue is empty or not */
	BUG_ON(queue->size < 0);
	BUG_ON(!queue->size && !list_empty(&queue->list));

	/* add entry at the back of the list */
	list_add_tail(&entry->_list, &queue->list);

	/* increment the queue size */
	queue->size++;

	/* enable (the update) interrupts again */
	spin_unlock_irqrestore(&queue->_slock, flags);

	return 0;
}

int
islpci_queue_get(islpci_queue_t *queue, islpci_qentry_t **entry)
{
	islpci_qentry_t *first = 0;
	unsigned long flags;

	BUG_ON(entry == 0);

	/* enter critical section */
	spin_lock_irqsave(&queue->_slock, flags);

	/* check whether the queue is empty or not */
	BUG_ON(queue->size < 0);
	if (queue->size == 0) {
		/* queue is empty, nothing to get */
		BUG_ON(!list_empty(&queue->list));
		spin_unlock_irqrestore(&queue->_slock, flags);
		*entry = 0;
		return -1;
	}

	/* the queue is not empty, get the entry at the start */
	BUG_ON(list_empty(&queue->list));
	first = list_entry(queue->list.next, islpci_qentry_t, _list);
	list_del(&first->_list);

	/* decrement the queue size */
	queue->size--;
	if (!queue->size)	/* we became empty */
		BUG_ON(!list_empty(&queue->list));

	/* we're done modifying the queue data structures */
	spin_unlock_irqrestore(&queue->_slock, flags);

	*entry = first;

	return 0;
}

int
islpci_queue_size(islpci_queue_t *queue)
{
	int qsize;
	unsigned long flags;

	/* enter critical section */
	spin_lock_irqsave(&queue->_slock, flags);
	qsize = queue->size;
	BUG_ON(!qsize && !list_empty(&queue->list));
	/* leave critical section */
	spin_unlock_irqrestore(&queue->_slock, flags);

	return qsize;
}

/******************************************************************************
    Device Interface functions
******************************************************************************/
int
islpci_mgt_transmit(islpci_private *priv)
{
	int fragments, counter;
	volatile isl38xx_control_block *control_block = priv->control_block;
	u32 index;
	islpci_qentry_t *entry;
	unsigned long flags;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_FUNCTION_CALLS, "islpci_mgt_transmit \n");
#endif

	/* check whether there is a pending response in the receive queue which
	 * needs to be handled first before applying a new frame */
	if (isl38xx_in_queue(control_block, ISL38XX_CB_RX_MGMTQ)) {
		/* receive queue not empty */
#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_TRACING, "receive queue not empty\n");
#endif
		return 0;
	}

	/* read the number of entries in the transmit queue and check whether the
	 * transmit queue is empty for receiving the next pimfor frame */
	if (isl38xx_in_queue(control_block, ISL38XX_CB_TX_MGMTQ)) {
		/* transmit queue not empty */
#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_TRACING, "transmit queue not empty\n");
#endif
		/* trigger the device */
		islpci_trigger(priv);

		return 0;
	}

	/* first check whether a clean up should take place on the shadow queue */
	while (priv->index_mgmt_tx > 0) {
		/* free the first entry in the shadow queue which
		 * contains the PIMFOR frame transmitted to the device */
		if (!islpci_queue_get(&priv->mgmt_tx_shadowq, &entry)) {
			/* decrement the real management index */
			priv->index_mgmt_tx--;
			islpci_queue_put(&priv->mgmt_tx_freeq, entry);
		} else {
			BUG();
		}
	}

	/* enter critical section */
	spin_lock_irqsave(&priv->mgmt_tx_shadowq._slock, flags);

	BUG_ON(priv->index_mgmt_tx > 0);

	if (!list_empty(&priv->mgmt_tx_shadowq.list)) {
		/* there is at least one frame in the shadowq, get its fragments count */
		entry =
		    list_entry(priv->mgmt_tx_shadowq.list.next, islpci_qentry_t,
			       _list);
		fragments = entry->fragments;

		/* transmit the fragment(s) by placing them in the transmit queue */
		for (counter = 0; counter < fragments; counter++) {
			volatile isl38xx_fragment *fragment;

			/* determine the fragment flags depending on the nr of fragments */
			u16 fragment_flags =
			    fragments - counter > 1 ? FRAGMENT_FLAG_MF : 0;

			BUG_ON(&entry->_list == &priv->mgmt_tx_shadowq.list);

			/* get a pointer to the destination fragment using the frame counter
			 * and link it with the free queue entry data block */
			index =
			    le32_to_cpu(control_block->
					driver_curr_frag[ISL38XX_CB_TX_MGMTQ]);
			index %= ISL38XX_CB_MGMT_QSIZE;
			fragment = &control_block->tx_data_mgmt[index];

#if VERBOSE > SHOW_ERROR_MESSAGES
			DEBUG(SHOW_TRACING, "Entry address %p\n", entry);
			DEBUG(SHOW_TRACING, "Index in transmit queue %i\n",
			      index);
			DEBUG(SHOW_TRACING, "Fragment address %p\n", fragment);
#endif

			fragment->address = cpu_to_le32(entry->dev_address);
			fragment->size = cpu_to_le16(entry->size);
			fragment->flags = cpu_to_le16(fragment_flags);

			/* increment the index pointer, wrapping is not done on queue but
			 * on variable type range */
			add_le32p((u32 *) & control_block->
				  driver_curr_frag[ISL38XX_CB_TX_MGMTQ], 1);

			/* increment the queue entry pointer */
			entry =
			    list_entry(entry->_list.next, islpci_qentry_t,
				       _list);
		}		/* for */

#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_TRACING,
		      "Frame with [%i] fragments added to transmit queue \n",
		      fragments);
#endif

		/* increment the read management queue index */
		priv->index_mgmt_tx += fragments;
	}

	/* if */
	/* done with critical section */
	spin_unlock_irqrestore(&priv->mgmt_tx_shadowq._slock, flags);

	/* trigger the device */
	islpci_trigger(priv);

	return 0;
}

#if 0
       /* ==================================================== */
if (priv->resetdevice == 1) {
	DEBUG(SHOW_FUNCTION_CALLS, "Reset Device flag set, quit IRQ handler\n");

	DEBUG(SHOW_QUEUE_INDEXES,
	      "CB drv Qs: [%i][%i][%i][%i][%i][%i]\n",
	      le32_to_cpu(priv->control_block->driver_curr_frag[0]),
	      le32_to_cpu(priv->control_block->driver_curr_frag[1]),
	      le32_to_cpu(priv->control_block->driver_curr_frag[2]),
	      le32_to_cpu(priv->control_block->driver_curr_frag[3]),
	      le32_to_cpu(priv->control_block->driver_curr_frag[4]),
	      le32_to_cpu(priv->control_block->driver_curr_frag[5])
	    );

	DEBUG(SHOW_QUEUE_INDEXES,
	      "CB dev Qs: [%i][%i][%i][%i][%i][%i]\n",
	      le32_to_cpu(priv->control_block->device_curr_frag[0]),
	      le32_to_cpu(priv->control_block->device_curr_frag[1]),
	      le32_to_cpu(priv->control_block->device_curr_frag[2]),
	      le32_to_cpu(priv->control_block->device_curr_frag[3]),
	      le32_to_cpu(priv->control_block->device_curr_frag[4]),
	      le32_to_cpu(priv->control_block->device_curr_frag[5])
	    );

	return 0;
}
	/* ==================================================== */
#endif

int
islpci_mgt_queue(islpci_private *priv,
		 int operation, unsigned long oid,
		 int flags, void *data, int length, int use_tunnel_oid)
{
	int fragments, counter;
	u32 source, destination;
	islpci_qentry_t *entry;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_FUNCTION_CALLS, "islpci_mgt_queue \n");
#endif

	/* determine the amount of fragments needed for the frame in case of a SET */
	if (use_tunnel_oid) {
		fragments = (length + 2 * PIMFOR_HEADER_SIZE) / MGMT_FRAME_SIZE;
		if ((length + 2 * PIMFOR_HEADER_SIZE) % MGMT_FRAME_SIZE != 0)
			fragments++;
	} else {
		fragments = (length + PIMFOR_HEADER_SIZE) / MGMT_FRAME_SIZE;
		if ((length + PIMFOR_HEADER_SIZE) % MGMT_FRAME_SIZE != 0)
			fragments++;
	}

	/* check whether data is actually written to the device */
	if (operation == PIMFOR_OP_GET) {
		/* operation is get, no data is actually written, reset the fragments
		 * counter to send a single fragment */
		fragments = 1;
	}
	/* check whether the number of fragments doesn't exceed the maximal possible
	 * amount of frames in the transmit queue */
	if (fragments > ISL38XX_CB_MGMT_QSIZE) {
		/* Oops, the fragment count exceeds the maximal value */
		DEBUG(SHOW_ERROR_MESSAGES,
		      "Error: Frame with [%i] fragments is too " "large \n",
		      fragments);
		return -1;
	}
	/* check whether there is enough room for the fragment(s) in the freeq */
	if (islpci_queue_size(&priv->mgmt_tx_freeq) < fragments) {
		/* Oops, the management freeq does not have enough entries */
		DEBUG(SHOW_ERROR_MESSAGES,
		      "Error: Management FreeQ not enough entries "
		      "for [%i] fragments \n", fragments);
		return -1;
	}
	/* queue the frame(s) in the transmit shadow queue */
	for (counter = 0, source = (u32) data; counter < fragments; counter++) {
		u16 fragment_size = 0;

		/* get an entry from the free queue for hosting the frame */
		islpci_queue_get(&priv->mgmt_tx_freeq, &entry);

		/* set the destination address to the entry data block for building
		 * the PIMFOR frame there */
		destination = entry->host_address;

#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_TRACING, "Entry address %p\n", entry);
		DEBUG(SHOW_TRACING, "Destination address %x\n", destination);
		DEBUG(SHOW_TRACING, "Destination for device %x\n",
		      entry->dev_address);
#endif

		/* place the PIMFOR header in the first fragment */
		if (counter == 0) {
			/* check whether the frame should have a tunnel oid */
			if (use_tunnel_oid) {
				/* start with a tunnel oid header */
				pimfor_encode_header(PIMFOR_OP_SET,
						     OID_INL_TUNNEL,
						     PIMFOR_DEV_ID_MHLI_MIB,
						     flags,
						     length +
						     PIMFOR_HEADER_SIZE,
						     (pimfor_header_t *)
						     destination);

				/* set the fragment size and increment the destin pointer */
				fragment_size += PIMFOR_HEADER_SIZE;
				destination += PIMFOR_HEADER_SIZE;
			}
			/* create the header directly in the fragment data area */
			pimfor_encode_header(operation, oid,
					     PIMFOR_DEV_ID_MHLI_MIB, flags,
					     length,
					     (pimfor_header_t *) destination);

#if VERBOSE > SHOW_ERROR_MESSAGES
			/* display the buffer contents for debugging */
			display_buffer((char *) destination, 12);
#endif

			/* set the fragment size and increment the destin pointer */
			fragment_size += PIMFOR_HEADER_SIZE;
			destination += PIMFOR_HEADER_SIZE;
		}
		/* add the data to the fragment  */
		if (length != 0 && source != 0) {
			int size =
			    (length >
			     (MGMT_FRAME_SIZE -
			      fragment_size)) ? MGMT_FRAME_SIZE -
			    fragment_size : length;
			memcpy((void *) destination, (void *) source, size);

			/* set the fragment size and flag members */
			fragment_size += size;

#if VERBOSE > SHOW_ERROR_MESSAGES
			/* display the buffer contents for debugging */
			display_buffer((char *) destination, 12);
#endif

			/* update the source pointer and the data length */
			source += size;
			length -= size;
		}
		/* save the number of fragments still to go in the frame together with
		 * the fragment size */
		entry->fragments = fragments - counter;
		entry->size = fragment_size;

		/* store the entry in the tx shadow queue */
		islpci_queue_put(&priv->mgmt_tx_shadowq, entry);
	}

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING,
	      "Frame with [%i] fragments added to shadow tx queue \n",
	      fragments);
#endif

	return 0;
}

int
islpci_mgt_receive(islpci_private *priv)
{
	struct net_device *ndev = priv->ndev;
	volatile isl38xx_control_block *control_block = priv->control_block;
	pimfor_header_t *received_header;
	char *received_data;
	int device_id, operation, flags;
	u32 oid, length;
	islpci_qentry_t *entry, *new_entry;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_FUNCTION_CALLS, "islpci_mgt_receive \n");
#endif

	/* get the received response from the shadow queue */
	if (islpci_queue_get(&priv->mgmt_rx_shadowq, &entry)) {
		printk(KERN_DEBUG
		       "%s: rx shadowq empty in islpci_mgt_receive\n",
		       ndev->name);
		return -1;
	}
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "Entry address 0x%p\n", entry);
	DEBUG(SHOW_TRACING, "Entry device address 0x%x\n", entry->dev_address);
#endif

	/* determine how the process the fragment depending on the flags in the
	 * PIMFOR header, check for tunnel oid PIMFOR frames */
	received_header = (pimfor_header_t *) entry->host_address;
	if (pimfor_decode_header(received_header, &operation, &oid, &device_id,
				 &flags, &length), oid == OID_INL_TUNNEL) {
#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_PIMFOR_FRAMES,
		      "PIMFOR: Tunnel OID frame received \n");
#endif

		/* PMFOR frame in fragment has the tunnel oid, redo the header decode */
		received_header = (pimfor_header_t *) (entry->host_address +
						       PIMFOR_HEADER_SIZE);
		pimfor_decode_header(received_header, &operation, &oid,
				     &device_id, &flags, &length);
	}

	/* The device ID from the pimfor packet received from the MVC is always 0
	 * This is not the device_id we want to give to the applications above
	 * so lets get the device_id using the private net_local structure
	 */
	device_id = priv->ndev->ifindex;

	received_data = (char *) received_header + PIMFOR_HEADER_SIZE;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_PIMFOR_FRAMES,
	      "PIMFOR: op %i, oid 0x%lx, device %i, flags 0x%x length 0x%x \n",
	      operation, oid, device_id, flags, length);

	/* display the buffer contents for debugging */
	display_buffer((char *) received_header, sizeof (pimfor_header_t));
	display_buffer(((char *) received_header) + PIMFOR_HEADER_SIZE, length);
#endif

	/* PIMFOR frame created by this driver, handle its contents */
	if ((oid >= DOT11_OID_DEAUTHENTICATE) &&
	    (oid <= DOT11_OID_REASSOCIATEEX)) {
		/* received an mlme reply or trap message */
		islpci_mgt_mlme(oid, received_data);
	}

	/* perform a check on the freeq whether the entry can be
	 * queued to other queues or must be hold to make sure there
	 * are enough */
	if (!islpci_queue_get(&priv->mgmt_rx_freeq, &new_entry)) {
		/* succeeded in getting a new entry from the freeq
		 * refresh the device mgmt queue entry */
		control_block->rx_data_mgmt[priv->index_mgmt_rx].address =
		    cpu_to_le32(new_entry->dev_address);

		/* put the new entry in the mgmt rx shadow queue */
		islpci_queue_put(&priv->mgmt_rx_shadowq, new_entry);

		/* deliver the held entry to the appropriate queue */
		if (flags & PIMFOR_FLAG_APPLIC_ORIGIN) {
			/* PIMFOR frame originated by higher
			 * application, send the entire fragment up
			 * without processing */

			/* send the entry to the PIMFOR receive queue
			   for further processing */
			islpci_queue_put(&priv->pimfor_rx_queue, entry);
		} else {
			/* check whether the frame concerns a response, error or trap */
			if (operation == PIMFOR_OP_TRAP) {
				printk(KERN_DEBUG
				       "TRAP: oid 0x%x, device %i, flags 0x%x length %i\n",
				       oid, device_id, flags, length);

			/* Add the trap to the queue and process it out of interrupt
			 * context. The entry will be freed there. */
			islpci_queue_put(&priv->trap_rx_queue, entry);
			schedule_work(&priv->trap_work);
			} else {
				/* received either a repsonse or an
				 * error send the entry to the IOCTL
				 * receive q for further processing */
				islpci_queue_put(&priv->ioctl_rx_queue, entry);

				/* signal the waiting procs that a
				   response has been received */
				wake_up_interruptible(&priv->
						      ioctl_reader_wqueue);

#if VERBOSE > SHOW_ERROR_MESSAGES
				DEBUG(SHOW_TRACING,
				      "Wake up interruptible Mgmt Queue\n");
#endif
			}
		}
	} else {
		/* not succeeded, use the held one and drop its contents
		 * refresh the device mgmt queue entry */
		control_block->rx_data_mgmt[priv->index_mgmt_rx].address =
		    cpu_to_le32(entry->dev_address);

		/* put the old entry back in the mgmt rx shadow queue */
		islpci_queue_put(&priv->mgmt_rx_shadowq, entry);
	}

	/* increment the driver read pointer */
	add_le32p((u32 *) & control_block->
		  driver_curr_frag[ISL38XX_CB_RX_MGMTQ], 1);

	/* trigger the device */
	islpci_trigger(priv);

	/* increment and wrap the read index for the rx management queue */
	priv->index_mgmt_rx = (priv->index_mgmt_rx + 1) % ISL38XX_CB_MGMT_QSIZE;

	return 0;
}

int
islpci_mgt_response(islpci_private *priv, long oid, int *oper,
		    void **data, long *dlen, islpci_qentry_t **qentry)
{
	const long wait_cycle_jiffies = (ISL38XX_WAIT_CYCLE * 10 * HZ) / 1000;

	long timeout_jiffies_left =
	    ISL38XX_MAX_WAIT_CYCLES * wait_cycle_jiffies;

	/* check for other processes already waiting for response */

	/* NOTE: this may cause to flush entries which have not been
	 * waited for yet, but are going to!  this seems to be a hack
	 * to flush the 18 replies created by the init routine */

	/* decrement the negative process counter and return it's
	   previous value */
	if (atomic_dec_and_test(&priv->ioctl_reader_count) == 0) {
		islpci_qentry_t *entry;
		int count = 0;

		/* no process in the wait queue, flush the queue
		 * flush each entry remaining in the queue */
		while (!islpci_queue_get(&priv->ioctl_rx_queue, &entry))
			islpci_queue_put(&priv->mgmt_rx_freeq, entry), count++;

		if (count)
			printk(KERN_WARNING
			       "islpci_mgt_response(): flushed %d entries\n",
			       count);
	}
	/* trigger the transmission of the management shadowq before
	   going to sleep */
	islpci_mgt_transmit(priv);

	/* set *qentry to 0, in order to catch bad calls to
	   islpci_mgt_release() */
	*qentry = 0;

	while (timeout_jiffies_left > 0) {
		int q_size;

		/* go to sleep with the alarm bells on sharp */
		timeout_jiffies_left -= wait_cycle_jiffies;
		timeout_jiffies_left +=
		    interruptible_sleep_on_timeout(&priv->ioctl_reader_wqueue,
						   wait_cycle_jiffies);

		/* neglect the jiffies spent in down_interruptible()... */
		if (down_interruptible(&priv->ioctl_reader_sem)) {
			/* signal! */
			printk(KERN_INFO
			       "signal while waiting for sem in islpci_mgt_response()\n");
			atomic_inc(&priv->ioctl_reader_count);
			return -ERESTARTSYS;
		}
		/* check the IOCTL receive queue for responses */
		if ((q_size = islpci_queue_size(&priv->ioctl_rx_queue)) != 0) {
			int counter;

			/* something in the queue, check every entry in the queue */
			for (counter = 0; counter < q_size; counter++) {
				pimfor_header_t *received_header;
				int device_id, operation, flags;
				u32 rlength, received_oid = 0;
				islpci_qentry_t *entry = 0;

				/* get the entry from the queue and read the header */
				if (islpci_queue_get
				    (&priv->ioctl_rx_queue, &entry))
					BUG();	/* should not happen since we are the only consumer */

				received_header =
				    (pimfor_header_t *) entry->host_address;

				/* check whether the oid response is for me */
				pimfor_decode_header(received_header,
						     &operation, &received_oid,
						     &device_id, &flags,
						     &rlength);

				if (oid == received_oid) {
					/* found my response, return it all */
					*data =
					    (void *) received_header +
					    PIMFOR_HEADER_SIZE;
					*qentry = entry;	/* the caller now owns the entry! */
					*dlen = rlength;
					*oper = operation;

					up(&priv->ioctl_reader_sem);
					atomic_inc(&priv->ioctl_reader_count);
					return 0;
				}
				/* put the entry back in the queue */
				islpci_queue_put(&priv->ioctl_rx_queue, entry);
			}

			/* checked all entries in the queue without success
			 * return -ETIMEDOUT;
			 * let's wait the full timeout */
			printk(KERN_INFO
			       "islpci_mgt_response(): entry in queue not found yet, retrying (jiffies left = %ld\n",
			       timeout_jiffies_left);
		} else {
			printk(KERN_INFO
			       "islpci_mgt_response(): queue empty, retrying (jiffies left = %ld\n",
			       timeout_jiffies_left);

			/* try another wakup on the card and wait for response
			 * when the device is sleeping */
			islpci_trigger(priv);
		}

		/* unlock semaphore */
		up(&priv->ioctl_reader_sem);
	}			/* while */

	/* seems we timed out (~1sec) */
	atomic_inc(&priv->ioctl_reader_count);

	return -ETIMEDOUT;
}

void
islpci_mgt_release(islpci_private *priv, islpci_qentry_t *entry)
{
	BUG_ON(entry == 0);

	islpci_queue_put(&priv->mgmt_rx_freeq, entry);
}

int
islpci_mgt_mlme(unsigned long oid, char *received_data)
{
	struct obj_mlme *mlme = (struct obj_mlme *) received_data;

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRAPS, "mlme: oid %lx \n", oid);
	DEBUG(SHOW_TRAPS, "mlme: address %02x:%02x:%02x:%02x:%02x:%02x\n",
	      mlme->address[0] & 0xFF, mlme->address[1] & 0xFF,
	      mlme->address[2] & 0xFF, mlme->address[3] & 0xFF,
	      mlme->address[4] & 0xFF, mlme->address[5] & 0xFF);
	DEBUG(SHOW_TRAPS, "mlme: id %x, state %x, code %x \n", mlme->id,
	      mlme->state, mlme->code);
#endif

	switch (oid) {
	case DOT11_OID_AUTHENTICATE:
		/* check whether a station is authenticating */
		if (mlme->state == DOT11_STATE_AUTHING) {
			/* STA is requesting to to be authenticated */

			/* check the code of the object and handle it */

			/* generate a response to the station
			 * positive: generate an Authenticate Confirm frame
			 * negative: generate a Deauthenicate Request frame */
		} else {
			/* notification that a STA has authenticated */

			/* update the STA list to be maintained in the driver */
		}
		break;

	case DOT11_OID_DEAUTHENTICATE:
		/* read the STA address from the mlme object structure */

		/* check the code of the object and handle it */

		/* update the STA list to be maintained in the driver */

		break;

	case DOT11_OID_ASSOCIATE:
		/* check whether a station is associating */
		if (mlme->state == DOT11_STATE_ASSOCING) {
			/* STA is requesting to to be authenticated */

			/* check the code of the object and handle it */

			/* generate a response to the station
			 * positive: generate an Associate Confirm frame
			 * negative: generate a Disassociate Request frame */
		} else {
			/* notification that a STA has authenticated */

			/* update the STA list to be maintained in the driver */
		}

		break;

	case DOT11_OID_DISASSOCIATE:
		/* read the STA address from the mlme object structure */

		/* update the STA list to be maintained in the driver */

		break;

	case DOT11_OID_BEACON:
		/* recevied a beacon indicate, do nothing */
#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_TRAPS, "TRAP: Beacon Indicate\n");
#endif
		break;

	case DOT11_OID_PROBE:
		/* recevied a probe indicate, do nothing */
#if VERBOSE > SHOW_ERROR_MESSAGES
		DEBUG(SHOW_TRAPS, "TRAP: Probe Indicate\n");
#endif
		break;

	default:
		/* oid is not a TRAP from the MLME */
		return 0;
	}

	return 0;

}

/******************************************************************************
        WDS Link Handlers
******************************************************************************/

#ifdef CONFIG_PRISM54_WDS
struct wds_link_msg {
	u8 mac[6];
	char name[32];
} __attribute__ ((packed));

int
islpci_wdslink_add_hndl(unsigned int dev_type, unsigned int dev_seq,
			unsigned int src_id, unsigned int src_seq,
			unsigned int op, unsigned int oid, void *data,
			unsigned long len)
{
	struct net_device *dev;
	struct wds_link_msg *wlmp;
	islpci_private *priv;
	char *datap;
	unsigned long dlen;
	islpci_qentry_t *entry = NULL;
	int err = 0;
	int operation;

	if (len != sizeof (struct wds_link_msg))
		return -1;

	if (!(dev = dev_get_by_index(dev_seq)))
		return -1;
	dev_put(dev);

	priv = dev->priv;

	wlmp = (struct wds_link_msg *) data;

	printk("islpci_wdslink_add_hndl(%x:%x:%x:%x:%x:%x)\n", wlmp->mac[0],
	       wlmp->mac[1], wlmp->mac[2], wlmp->mac[3], wlmp->mac[4],
	       wlmp->mac[5]);

	if (add_wds_link(dev, &priv->wdsp, wlmp->mac, wlmp->name) < 0) {
		printk("islpci_wdslink_add_hndl:add_wds_link\n");
	}

	islpci_mgt_queue(priv, op, oid, 0, wlmp->mac, 6, 0);
	if ((err = islpci_mgt_response(priv, oid, &operation,
				       (void **) &datap, &dlen, &entry)) < 0) {
		printk("islpci_mgt_response returned %d\n", err);
	}

	islpci_mgt_release(priv, entry);

	if ((mgt_response(src_id, src_seq, dev_type, dev_seq, operation,
			  oid, data, len)) < 0)
		return -1;

	return 0;
}

int
islpci_wdslink_del_hndl(unsigned int dev_type, unsigned int dev_seq,
			unsigned int src_id, unsigned int src_seq,
			unsigned int op, unsigned int oid, void *data,
			unsigned long len)
{
	struct net_device *dev;
	struct wds_link_msg *wlmp;
	islpci_private *priv;
	char *datap;
	unsigned long dlen;
	islpci_qentry_t *entry = NULL;
	int err = 0;
	int operation;

	if (len != sizeof (struct wds_link_msg))
		return -1;

	if (!(dev = dev_get_by_index(dev_seq)))
		return -1;
	dev_put(dev);

	priv = dev->priv;

	wlmp = (struct wds_link_msg *) data;

	printk("islpci_wdslink_del_hndl(%x:%x:%x:%x:%x:%x)\n", wlmp->mac[0],
	       wlmp->mac[1], wlmp->mac[2], wlmp->mac[3], wlmp->mac[4],
	       wlmp->mac[5]);

	if (del_wds_link(dev, &priv->wdsp, wlmp->mac, wlmp->name) < 0) {
		printk("islpci_wdslink_del_hndl:del_wds_link\n");
	}

	islpci_mgt_queue(priv, op, oid, 0, wlmp->mac, 6, 0);
	if ((err = islpci_mgt_response(priv, oid, &operation,
				       (void **) &datap, &dlen, &entry)) < 0) {
		printk("islpci_mgt_response returned %d\n", err);
	}

	islpci_mgt_release(priv, entry);

	if ((mgt_response(src_id, src_seq, dev_type, dev_seq, operation,
			  oid, data, len)) < 0)
		return -1;

	return 0;
}
#endif				/* CONFIG_PRISM54_WDS */

void
islpci_parse_trap(islpci_qentry_t *entry, u32 *oid, char **data)
{
	pimfor_header_t *header;
	int device_id, operation, flags;
	u32 length;

	header = (pimfor_header_t *) entry->host_address;
	pimfor_decode_header(header, &operation, oid, &device_id, &flags, &length);
	*data = (char *) header + PIMFOR_HEADER_SIZE;
	BUG_ON(operation != PIMFOR_OP_TRAP);
}

